import io
import os
from ..cadnano import app
from ..model.document import Document
from ..model.io.decoder import decode
from ..model.io.encoder import encode
from ..views.documentwindow import DocumentWindow
from ..views import styles
import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), ['QDir', 'QFileInfo', 'QRect',
                                        'QSettings',
                                        'QSize', 'Qt'])
util.qtWrapImport('QtGui', globals(), [
                                       'QIcon',
                                       'QKeySequence',
                                       'QPainter'
                                       ])
util.qtWrapImport('QtWidgets', globals(), ['QApplication',
                                           'QDialog',
                                           'QDockWidget',
                                           'QFileDialog',
                                           'QGraphicsItem',
                                           'QMainWindow',
                                           'QMessageBox',
                                           'QStyleOptionGraphicsItem'])
util.qtWrapImport('QtSvg', globals(), ['QSvgGenerator'])


class DocumentController():
    """
    Connects UI buttons to their corresponding actions in the model.
    """
    ### INIT METHODS ###
    def __init__(self):
        """docstring for __init__"""
        # initialize variables
        self._document = Document()
        self._document.setController(self)
        self._activePart = None
        self._filename = None
        self._fileOpenPath = None  # will be set in _readSettings
        self._hasNoAssociatedFile = True
        self._pathViewInstance = None
        self._sliceViewInstance = None
        self._undoStack = None
        self.win = None
        self.fileopendialog = None
        self.filesavedialog = None

        self.settings = QSettings()
        self._readSettings()

        QDir.addSearchPath('icons', 'ui/mainwindow/images/')

        # call other init methods
        self._initWindow()
        if app().isInMaya():
            self._initMaya()
        app().documentControllers.add(self)

    def _initWindow(self):
        """docstring for initWindow"""
        self.win = DocumentWindow(docCtrlr=self)
        self.win.setWindowIcon(QIcon('icons:cadnano2-app-icon.png'))
        app().documentWindowWasCreatedSignal.emit(self._document, self.win)
        self._connectWindowSignalsToSelf()
        self.win.show()

    def _initMaya(self):
        """
        Initialize Maya-related state. Delete Maya nodes if there
        is an old document left over from the same session. Set up
        the Maya window.
        """
        # There will only be one document
        if (app().activeDocument and app().activeDocument.win and
                                not app().activeDocument.win.close()):
            return
        del app().activeDocument
        app().activeDocument = self

        import maya.OpenMayaUI as OpenMayaUI
        import sip
        ptr = OpenMayaUI.MQtUtil.mainWindow()
        mayaWin = sip.wrapinstance(int(ptr), QMainWindow)
        self.windock = QDockWidget("cadnano")
        self.windock.setFeatures(QDockWidget.DockWidgetMovable
                                 | QDockWidget.DockWidgetFloatable)
        self.windock.setAllowedAreas(Qt.LeftDockWidgetArea
                                     | Qt.RightDockWidgetArea)
        self.windock.setWidget(self.win)
        mayaWin.addDockWidget(Qt.DockWidgetArea(Qt.LeftDockWidgetArea),
                                self.windock)
        self.windock.setVisible(True)

    def destroyDC(self):
        self.disconnectSignalsToSelf()
        if self.win is not None:
            self.win.destroyWin()
            self.win = None
    # end def

    def disconnectSignalsToSelf(self):
        win = self.win
        if win is not None:
            win.actionNew.triggered.disconnect(self.actionNewSlot)
            win.actionOpen.triggered.disconnect(self.actionOpenSlot)
            win.actionDrop.triggered.disconnect(self.actionDropSlot)
            win.actionClose.triggered.disconnect(self.actionCloseSlot)
            win.actionSave.triggered.disconnect(self.actionSaveSlot)
            win.actionSave_As.triggered.disconnect(self.actionSaveAsSlot)
            win.actionSVG.triggered.disconnect(self.actionSVGSlot)
            win.actionAutoStaple.triggered.disconnect(self.actionAutostapleSlot)
            win.actionExportStaples.triggered.disconnect(self.actionExportStaplesSlot)
            win.actionPreferences.triggered.disconnect(self.actionPrefsSlot)
            win.actionModify.triggered.disconnect(self.actionModifySlot)
            win.actionNewHoneycombPart.triggered.disconnect(self.actionAddHoneycombPartSlot)
            win.actionNewSquarePart.triggered.disconnect(self.actionAddSquarePartSlot)
            win.closeEvent = self.windowCloseEventHandler
            win.actionAbout.triggered.disconnect(self.actionAboutSlot)
            win.actionCadnanoWebsite.triggered.disconnect(self.actionCadnanoWebsiteSlot)
            win.actionFeedback.triggered.disconnect(self.actionFeedbackSlot)
            win.actionFilterHandle.triggered.disconnect(self.actionFilterHandleSlot)
            win.actionFilterEndpoint.triggered.disconnect(self.actionFilterEndpointSlot)
            win.actionFilterStrand.triggered.disconnect(self.actionFilterStrandSlot)
            win.actionFilterXover.triggered.disconnect(self.actionFilterXoverSlot)
            win.actionFilterScaf.triggered.disconnect(self.actionFilterScafSlot)
            win.actionFilterStap.triggered.disconnect(self.actionFilterStapSlot)
            win.actionRenumber.triggered.disconnect(self.actionRenumberSlot)
    # end def

    def _connectWindowSignalsToSelf(self):
        """This method serves to group all the signal & slot connections
        made by DocumentController"""
        self.win.actionNew.triggered.connect(self.actionNewSlot)
        self.win.actionOpen.triggered.connect(self.actionOpenSlot)
        self.win.actionDrop.triggered.connect(self.actionDropSlot)
        self.win.actionClose.triggered.connect(self.actionCloseSlot)
        self.win.actionSave.triggered.connect(self.actionSaveSlot)
        self.win.actionSave_As.triggered.connect(self.actionSaveAsSlot)
        self.win.actionSVG.triggered.connect(self.actionSVGSlot)
        self.win.actionAutoStaple.triggered.connect(self.actionAutostapleSlot)
        self.win.actionExportStaples.triggered.connect(self.actionExportStaplesSlot)
        self.win.actionPreferences.triggered.connect(self.actionPrefsSlot)
        self.win.actionModify.triggered.connect(self.actionModifySlot)
        self.win.actionNewHoneycombPart.triggered.connect(self.actionAddHoneycombPartSlot)
        self.win.actionNewSquarePart.triggered.connect(self.actionAddSquarePartSlot)
        self.win.closeEvent = self.windowCloseEventHandler
        self.win.actionAbout.triggered.connect(self.actionAboutSlot)
        self.win.actionCadnanoWebsite.triggered.connect(self.actionCadnanoWebsiteSlot)
        self.win.actionFeedback.triggered.connect(self.actionFeedbackSlot)
        self.win.actionFilterHandle.triggered.connect(self.actionFilterHandleSlot)
        self.win.actionFilterEndpoint.triggered.connect(self.actionFilterEndpointSlot)
        self.win.actionFilterStrand.triggered.connect(self.actionFilterStrandSlot)
        self.win.actionFilterXover.triggered.connect(self.actionFilterXoverSlot)
        self.win.actionFilterScaf.triggered.connect(self.actionFilterScafSlot)
        self.win.actionFilterStap.triggered.connect(self.actionFilterStapSlot)
        self.win.actionRenumber.triggered.connect(self.actionRenumberSlot)


    ### SLOTS ###
    def undoStackCleanChangedSlot(self):
        """The title changes to include [*] on modification."""
        self.win.setWindowModified(not self.undoStack().isClean())
        self.win.setWindowTitle(self.documentTitle())

    def actionAboutSlot(self):
        """Displays the about cadnano dialog."""
        from cadnano2.ui.dialogs.ui_about import Ui_About
        dialog = QDialog()
        dialogAbout = Ui_About()  # reusing this dialog, should rename
        dialog.setStyleSheet("QDialog { background-image: url(ui/dialogs/images/cadnano2-about.png); background-repeat: none; }")
        dialogAbout.setupUi(dialog)
        dialog.exec()

    filterList = ["strand", "endpoint", "xover", "virtualHelix"]
    def actionFilterHandleSlot(self):
        """Disables all other selection filters when active."""
        fH = self.win.actionFilterHandle
        fE = self.win.actionFilterEndpoint
        fS = self.win.actionFilterStrand
        fX = self.win.actionFilterXover
        fH.setChecked(True)
        if fE.isChecked():
            fE.setChecked(False)
        if fS.isChecked():
            fS.setChecked(False)
        if fX.isChecked():
            fX.setChecked(False)
        self._document.documentSelectionFilterChangedSignal.emit(["virtualHelix"])

    def actionFilterEndpointSlot(self):
        """
        Disables handle filters when activated.
        Remains checked if no other item-type filter is active.
        """
        fH = self.win.actionFilterHandle
        fE = self.win.actionFilterEndpoint
        fS = self.win.actionFilterStrand
        fX = self.win.actionFilterXover
        if fH.isChecked():
            fH.setChecked(False)
        if not fS.isChecked() and not fX.isChecked():
            fE.setChecked(True)
        self._strandFilterUpdate()
    # end def

    def actionFilterStrandSlot(self):
        """
        Disables handle filters when activated.
        Remains checked if no other item-type filter is active.
        """
        fH = self.win.actionFilterHandle
        fE = self.win.actionFilterEndpoint
        fS = self.win.actionFilterStrand
        fX = self.win.actionFilterXover
        if fH.isChecked():
            fH.setChecked(False)
        if not fE.isChecked() and not fX.isChecked():
            fS.setChecked(True)
        self._strandFilterUpdate()
    # end def

    def actionFilterXoverSlot(self):
        """
        Disables handle filters when activated.
        Remains checked if no other item-type filter is active.
        """
        fH = self.win.actionFilterHandle
        fE = self.win.actionFilterEndpoint
        fS = self.win.actionFilterStrand
        fX = self.win.actionFilterXover
        if fH.isChecked():
            fH.setChecked(False)
        if not fE.isChecked() and not fS.isChecked():
            fX.setChecked(True)
        self._strandFilterUpdate()
    # end def

    def actionFilterScafSlot(self):
        """Remains checked if no other strand-type filter is active."""
        fSc = self.win.actionFilterScaf
        fSt = self.win.actionFilterStap
        if not fSc.isChecked() and not fSt.isChecked():
            fSc.setChecked(True)
        self._strandFilterUpdate()

    def actionFilterStapSlot(self):
        """Remains checked if no other strand-type filter is active."""
        fSc = self.win.actionFilterScaf
        fSt = self.win.actionFilterStap
        if not fSc.isChecked() and not fSt.isChecked():
            fSt.setChecked(True)
        self._strandFilterUpdate()
    # end def

    def _strandFilterUpdate(self):
        win = self.win
        filterList = []
        if win.actionFilterEndpoint.isChecked():
            filterList.append("endpoint")
        if win.actionFilterStrand.isChecked():
            filterList.append("strand")
        if win.actionFilterXover.isChecked():
            filterList.append("xover")
        if win.actionFilterScaf.isChecked():
            filterList.append("scaffold")
        if win.actionFilterStap.isChecked():
            filterList.append("staple")
        self._document.documentSelectionFilterChangedSignal.emit(filterList)
    # end def

    def actionNewSlot(self):
        """
        1. If document is has no parts, do nothing.
        2. If document is dirty, call maybeSave and continue if it succeeds.
        3. Create a new document and swap it into the existing ctrlr/window.
        """
        # clear/reset the view!

        if len(self._document.parts()) == 0:
            return  # no parts
        if self.maybeSave() == False:
            return  # user canceled in maybe save
        else:  # user did not cancel
            if self.filesavedialog != None:
                self.filesavedialog.finished.connect(self.newClickedCallback)
            else:  # user did not save
                self.newClickedCallback()  # finalize new

    def actionOpenSlot(self):
        """
        1. If document is untouched, proceed to open dialog.
        2. If document is dirty, call maybesave and continue if it succeeds.
        Downstream, the file is selected in openAfterMaybeSave,
        and the selected file is actually opened in openAfterMaybeSaveCallback.
        """
        if self.maybeSave() == False:
            return  # user canceled in maybe save
        else:  # user did not cancel
            if hasattr(self, "filesavedialog"): # user did save
                if self.filesavedialog != None:
                    self.filesavedialog.finished.connect(self.openAfterMaybeSave)
                else:
                    self.openAfterMaybeSave()  # windows
            else:  # user did not save
                self.openAfterMaybeSave()  # finalize new

    def actionDropSlot(self):
        """
        1. If document is untouched, proceed to open dialog.
        2. If document is dirty, call maybesave and continue if it succeeds.
        Equivalent to actionOpenSlot() for Drag and Drop Event.
        """
        if self.maybeSave() is False:
            return  # user canceled in maybe save
        else:  # user did not cancel
            if hasattr(self, "filesavedialog"):  # user did save
                if self.filesavedialog is not None:
                    self.filesavedialog.finished.connect(self.openDropAfterMaybeSave)
                else:
                    self.openDropAfterMaybeSave()  # windows
            else:  # user did not save
                self.openDropAfterMaybeSave()  # finalize new

    def actionCloseSlot(self):
        """This will trigger a Window closeEvent."""
        if util.isWindows():
            #print "close win"
            if self.win is not None:
                self.win.close()
            if not app().isInMaya():
                #print "exit app"
                import sys
                sys.exit(1)

    def actionSaveSlot(self):
        """SaveAs if necessary, otherwise overwrite existing file."""
        if self._hasNoAssociatedFile:
            self.saveFileDialog()
            return
        self.writeDocumentToFile()

    def actionSaveAsSlot(self):
        """Open a save file dialog so user can choose a name."""
        self.saveFileDialog()

    def actionSVGSlot(self):
        """docstring for actionSVGSlot"""
        fname = os.path.basename(str(self.filename()))
        if fname == None:
            directory = "."
        else:
            directory = QFileInfo(fname).path()

        fdialog = QFileDialog(
                    self.win,
                    "%s - Save As" % QApplication.applicationName(),
                    directory,
                    "%s (*.svg)" % QApplication.applicationName())
        fdialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        fdialog.setWindowFlags(Qt.WindowType.Sheet)
        fdialog.setWindowModality(Qt.WindowModality.WindowModal)
        self.svgsavedialog = fdialog
        self.svgsavedialog.filesSelected.connect(self.saveSVGDialogCallback)
        fdialog.open()

    class DummyChild(QGraphicsItem):
        def boundingRect(self):
            return QRect(200, 200) # self.parentObject().boundingRect()
        def paint(self, painter, option, widget=None):
            pass

    def saveSVGDialogCallback(self, selected):
        if isinstance(selected, (list, tuple)):
            fname = selected[0]
        else:
            fname = selected
        if not fname or fname is None or os.path.isdir(fname):
            return False
        fname = str(fname)
        if not fname.lower().endswith(".svg"):
            fname += ".svg"
        if self.svgsavedialog != None:
            self.svgsavedialog.filesSelected.disconnect(self.saveSVGDialogCallback)
            del self.svgsavedialog  # prevents hang
            self.svgsavedialog = None

        generator = QSvgGenerator()
        generator.setFileName(fname)
        generator.setSize(QSize(200, 200))
        generator.setViewBox(QRect(0, 0, 2000, 2000))
        painter = QPainter()

        # Render through scene
        # painter.begin(generator)
        # self.win.pathscene.render(painter)
        # painter.end()

        # Render item-by-item
        painter = QPainter()
        styleOption = QStyleOptionGraphicsItem()
        q = [self.win.pathroot]
        painter.begin(generator)
        while q:
            graphicsItem = q.pop()
            transform = graphicsItem.itemTransform(self.win.sliceroot)[0]
            painter.setTransform(transform)
            if graphicsItem.isVisible():
                graphicsItem.paint(painter, styleOption, None)
                q.extend(graphicsItem.childItems())
        painter.end()

    def actionExportStaplesSlot(self):
        """
        Triggered by clicking Export Staples button. Opens a file dialog to
        determine where the staples should be saved. The callback is
        exportStaplesCallback which collects the staple sequences and exports
        the file.
        """
        # Validate that no staple oligos are loops.
        part = self.activePart()
        stapLoopOlgs = part.getStapleLoopOligos()
        if stapLoopOlgs:
            from ui.dialogs.ui_warning import Ui_Warning
            dialog = QDialog()
            dialogWarning = Ui_Warning()  # reusing this dialog, should rename
            dialog.setStyleSheet("QDialog { background-image: url(ui/dialogs/images/cadnano2-about.png); background-repeat: none; }")
            dialogWarning.setupUi(dialog)

            locs = ", ".join([o.locString() for o in stapLoopOlgs])
            msg = "Part contains staple loop(s) at %s.\n\nUse the break tool to introduce 5' & 3' ends before exporting. Loops have been colored red; use undo to revert." % locs
            dialogWarning.title.setText("Staple validation failed")
            dialogWarning.message.setText(msg)
            for o in stapLoopOlgs:
                o.applyColor(styles.stapColors[0].name())
            dialog.exec()
            return

        # Proceed with staple export.
        fname = self.filename()
        if fname == None:
            directory = "."
        else:
            directory = QFileInfo(fname).path()
        if util.isWindows():  # required for native looking file window
            fname = QFileDialog.getSaveFileName(
                            self.win,
                            "%s - Export As" % QApplication.applicationName(),
                            directory,
                            "(*.csv)")
            self.saveStaplesDialog = None
            self.exportStaplesCallback(fname)
        else:  # access through non-blocking callback
            fdialog = QFileDialog(
                            self.win,
                            "%s - Export As" % QApplication.applicationName(),
                            directory,
                            "(*.csv)")
            fdialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
            fdialog.setWindowFlags(Qt.WindowType.Sheet)
            fdialog.setWindowModality(Qt.WindowModality.WindowModal)
            self.saveStaplesDialog = fdialog
            self.saveStaplesDialog.filesSelected.connect(self.exportStaplesCallback)
            fdialog.open()
    # end def

    def actionPrefsSlot(self):
        app().prefsClicked()

    def actionAutostapleSlot(self):
        part = self.activePart()
        if part:
            self.win.pathGraphicsView.setViewportUpdateOn(False)
            part.autoStaple()
            self.win.pathGraphicsView.setViewportUpdateOn(True)

    def actionModifySlot(self):
        """
        Notifies that part root items that parts should respond to modifier
        selection signals.
        """
        # uncomment for debugging
        # isChecked = self.win.actionModify.isChecked()
        # self.win.pathroot.setModifyState(isChecked)
        # self.win.sliceroot.setModifyState(isChecked)
        if app().isInMaya():
            isChecked = self.win.actionModify.isChecked()
            self.win.pathroot.setModifyState(isChecked)
            self.win.sliceroot.setModifyState(isChecked)
            self.win.solidroot.setModifyState(isChecked)

    def actionAddHoneycombPartSlot(self):
        """docstring for actionAddHoneycombPartSlot"""
        part = self._document.addHoneycombPart()
        self.setActivePart(part)

    def actionAddSquarePartSlot(self):
        """docstring for actionAddSquarePartSlot"""
        part = self._document.addSquarePart()
        self.setActivePart(part)

    def actionRenumberSlot(self):
        coordList = self.win.pathroot.getSelectedPartOrderedVHList()
        part = self.activePart()
        part.renumber(coordList)
    # end def

    ### ACCESSORS ###
    def document(self):
        return self._document

    def window(self):
        return self.win

    def setDocument(self, doc):
        """
        Sets the controller's document, and informs the document that
        this is its controller.
        """
        self._document = doc
        doc.setController(self)

    def activePart(self):
        if self._activePart == None:
            self._activePart = self._document.selectedPart()
        return self._activePart

    def setActivePart(self, part):
        self._activePart = part

    def undoStack(self):
        return self._document.undoStack()

    ### PRIVATE SUPPORT METHODS ###
    def newDocument(self, doc=None, fname=None):
        """Creates a new Document, reusing the DocumentController."""
        self._document.resetViews()
        self._document.removeAllParts()  # clear out old parts
        self._document.undoStack().clear()  # reset undostack
        self._filename = fname if fname else "untitled.json"
        self._hasNoAssociatedFile = fname == None
        self._activePart = None
        self.win.setWindowTitle(self.documentTitle() + '[*]')

    def saveFileDialog(self):
        fname = self.filename()
        if fname == None:
            directory = "."
        else:
            directory = QFileInfo(fname).path()
        if util.isWindows():  # required for native looking file window
            fname = QFileDialog.getSaveFileName(
                            self.win,
                            "%s - Save As" % QApplication.applicationName(),
                            directory,
                            "%s (*.json)" % QApplication.applicationName())
            self.writeDocumentToFile(fname)
        else:  # access through non-blocking callback
            fdialog = QFileDialog(
                            self.win,
                            "%s - Save As" % QApplication.applicationName(),
                            directory,
                            "%s (*.json)" % QApplication.applicationName())
            fdialog.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
            fdialog.setWindowFlags(Qt.WindowType.Sheet)
            fdialog.setWindowModality(Qt.WindowModality.WindowModal)
            self.filesavedialog = fdialog
            self.filesavedialog.filesSelected.connect(
                                                self.saveFileDialogCallback)
            fdialog.open()
    # end def

    def _readSettings(self):
        self.settings.beginGroup("FileSystem")
        self._fileOpenPath = self.settings.value("openpath", QDir().homePath())
        self.settings.endGroup()

    def _writeFileOpenPath(self, path):
        """docstring for _writePath"""
        self._fileOpenPath = path
        self.settings.beginGroup("FileSystem")
        self.settings.setValue("openpath", path)
        self.settings.endGroup()

    ### SLOT CALLBACKS ###
    def actionNewSlotCallback(self):
        """
        Gets called on completion of filesavedialog after newClicked's
        maybeSave. Removes the dialog if necessary, but it was probably
        already removed by saveFileDialogCallback.
        """
        if self.filesavedialog != None:
            self.filesavedialog.finished.disconnect(self.actionNewSlotCallback)
            del self.filesavedialog  # prevents hang (?)
            self.filesavedialog = None
        self.newDocument()

    def exportStaplesCallback(self, selected):
        """Export all staple sequences to selected CSV file.

        Args:
            selected (Tuple, List or str): if a List or Tuple, the filename
            should be the first element
        """
        if isinstance(selected, (list, tuple)):
            fname = selected[0]
        else:
            fname = selected
        # Return if fname is '', None, or a directory path
        if not fname or fname is None or os.path.isdir(fname):
            return False
        # if not fname.lower().endswith(".txt"):
        #     fname += ".txt"
        if self.saveStaplesDialog is not None:
            self.saveStaplesDialog.filesSelected.disconnect(self.exportStaplesCallback)
            # manual garbage collection to prevent hang (in osx)
            del self.saveStaplesDialog
            self.saveStaplesDialog = None
        # write the file
        ap = self.activePart()
        if ap is not None:
            output = ap.getStapleSequences()
            with open(fname, 'w') as f:
                f.write(output)
    # end def

    def newClickedCallback(self):
        """
        Gets called on completion of filesavedialog after newClicked's
        maybeSave. Removes the dialog if necessary, but it was probably
        already removed by saveFileDialogCallback.
        """

        if self.filesavedialog != None:
            self.filesavedialog.finished.disconnect(self.newClickedCallback)
            del self.filesavedialog  # prevents hang (?)
            self.filesavedialog = None
        self.newDocument()

    def openAfterMaybeSaveCallback(self, selected):
        """
        Receives file selection info from the dialog created by
        openAfterMaybeSave, following user input.

        Extracts the file name and passes it to the decode method, which
        returns a new document doc, which is then set as the open document
        by newDocument. Calls finalizeImport and disconnects dialog signaling.
        """
        if isinstance(selected, (list, tuple)):
            fname = selected[0]
        else:
            fname = selected
        if fname is None or fname == '' or os.path.isdir(fname):
            return False
        if not os.path.exists(fname):
            return False
        self._writeFileOpenPath(os.path.dirname(fname))
        self.newDocument(fname=fname)

        with io.open(fname, 'r', encoding='utf-8') as fd:
            decode(self._document, fd.read())

        if hasattr(self, "filesavedialog"):  # user did save
            if self.fileopendialog is not None:
                self.fileopendialog.filesSelected.disconnect(self.openAfterMaybeSaveCallback)
            # manual garbage collection to prevent hang (in osx)
            del self.fileopendialog
            self.fileopendialog = None

    def saveFileDialogCallback(self, selected):
        """If the user chose to save, write to that file."""
        if isinstance(selected, (list, tuple)):
            fname = selected[0]
        else:
            fname = selected
        if fname is None or os.path.isdir(fname):
            return False
        if not fname.lower().endswith(".json"):
            fname += ".json"
        if self.filesavedialog is not None:
            self.filesavedialog.filesSelected.disconnect(self.saveFileDialogCallback)
            del self.filesavedialog  # prevents hang
            self.filesavedialog = None
        self.writeDocumentToFile(fname)
        self._writeFileOpenPath(os.path.dirname(fname))

    ### EVENT HANDLERS ###
    def windowCloseEventHandler(self, event):
        """Intercept close events when user attempts to close the window."""
        if self.maybeSave():
            event.accept()
            if app().isInMaya():
                self.windock.setVisible(False)
                del self.windock
                self.windock = None
            the_app = app()
            self.destroyDC()
            if the_app.documentControllers:
                the_app.destroyApp()
        else:
            event.ignore()
        self.actionCloseSlot()

    ### FILE INPUT ##
    def documentTitle(self):
        fname = os.path.basename(str(self.filename()))
        if not self.undoStack().isClean():
            fname += '[*]'
        return fname

    def filename(self):
        return self._filename

    def setFilename(self, proposedFName):
        if self._filename == proposedFName:
            return True
        self._filename = proposedFName
        self._hasNoAssociatedFile = False
        self.win.setWindowTitle(self.documentTitle())
        return True

    def openDropAfterMaybeSave(self):
        """
        This is the method that initiates file opening after a drag and Drop event.
        It is called by actionDropSlot.
        """
        fname = self.win._dropped_file
        if fname.endswith(".json"):
            self.openAfterMaybeSaveCallback(fname)
        else:
            print(f"Ignoring dropped file {fname}. Use a cadnano.json file.")

    def openAfterMaybeSave(self):
        """
        This is the method that initiates file opening. It is called by
        actionOpenSlot to spawn a QFileDialog and connect it to a callback
        method.
        """
        path = self._fileOpenPath
        if util.isWindows():  # required for native looking file window#"/",
            fname = QFileDialog.getOpenFileName(
                        None,
                        "Open Document", path,
                        "cadnano1 / cadnano2 Files (*.nno *.json *.cadnano)")
            self.filesavedialog = None
            self.openAfterMaybeSaveCallback(fname)
        else:  # access through non-blocking callback
            fdialog = QFileDialog(
                        self.win,
                        "Open Document",
                        path,
                        "cadnano1 / cadnano2 Files (*.nno *.json *.cadnano)")
            fdialog.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
            fdialog.setWindowFlags(Qt.WindowType.Sheet)
            fdialog.setWindowModality(Qt.WindowModality.WindowModal)
            self.fileopendialog = fdialog
            self.fileopendialog.filesSelected.connect(self.openAfterMaybeSaveCallback)
            fdialog.open()
    # end def

    ### FILE OUTPUT ###
    def maybeSave(self):
        """Save on quit, check if document changes have occured."""
        if app().dontAskAndJustDiscardUnsavedChanges:
            return True
        if not self.undoStack().isClean():    # document dirty?
            savebox = QMessageBox(QMessageBox.Icon.Question,   "Application",
                "The document has been modified.\nDo you want to save your changes?",
                QMessageBox.StandardButton.Save | QMessageBox.StandardButton.Discard | QMessageBox.StandardButton.Cancel,
                self.win,
                Qt.WindowType.Dialog | Qt.WindowType.MSWindowsFixedSizeDialogHint | Qt.WindowType.Sheet)
            savebox.setWindowModality(Qt.WindowModality.WindowModal)
            save = savebox.button(QMessageBox.StandardButton.Save)
            discard = savebox.button(QMessageBox.StandardButton.Discard)
            cancel = savebox.button(QMessageBox.StandardButton.Cancel)
            save.setShortcut("Ctrl+S")
            discard.setShortcut(QKeySequence("D,Ctrl+D"))
            cancel.setShortcut(QKeySequence("C,Ctrl+C,.,Ctrl+."))
            ret = savebox.exec()
            del savebox  # manual garbage collection to prevent hang (in osx)
            if ret == QMessageBox.StandardButton.Save:
                return self.actionSaveAsSlot()
            elif ret == QMessageBox.StandardButton.Cancel:
                return False
        return True

    def writeDocumentToFile(self, filename=None):
        helixOrderList = self.win.pathroot.getSelectedPartOrderedVHList()

        if helixOrderList == None:  # Do not attempt to save an empty design
            print("Cannot save empty document.")
            return False

        if filename == None:
            assert(not self._hasNoAssociatedFile)
            filename = self.filename()
        try:
            if util.isWindows() and isinstance(filename, (list,tuple)):
                filename = filename[0]
            with open(filename, 'w') as f:
                encode(self._document, helixOrderList, f)
        except IOError:
            flags = Qt.Dialog | Qt.MSWindowsFixedSizeDialogHint | Qt.WindowType.Sheet
            errorbox = QMessageBox(QMessageBox.Critical,
                                   "cadnano",
                                   "Could not write to '%s'." % filename,
                                   QMessageBox.Ok,
                                   self.win,
                                   flags)
            errorbox.setWindowModality(Qt.WindowModality.WindowModal)
            errorbox.open()
            return False
        self.undoStack().setClean()
        self.setFilename(filename)
        return True

    def actionCadnanoWebsiteSlot(self):
        import webbrowser
        webbrowser.open("http://cadnano.org/")

    def actionFeedbackSlot(self):
        import webbrowser
        webbrowser.open("http://cadnano.org/feedback")



import sys
from .abstractpathtool import AbstractPathTool
import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), [])
util.qtWrapImport('QtGui', globals(), [])


class PaintTool(AbstractPathTool):
    """
    Handles visibility and color cycling for the paint tool.
    """
    def __init__(self, controller):
        super(PaintTool, self).__init__(controller)
        self._isMacrod = False

    def __repr__(self):
        return "paintTool"  # first letter should be lowercase

    def setActive(self, willBeActive):
        """Show the ColorPicker widget when active, hide when inactive."""
        if willBeActive:
            self._window.pathColorPanel.show()
        else:
            self._window.pathColorPanel.hide()
            self._window.pathColorPanel.prevColor()

    def widgetClicked(self):
        """Cycle through colors on 'p' keypress"""
        self._window.pathColorPanel.nextColor()
        
    def customMouseRelease(self, event):
        if self._isMacrod:
            self._isMacrod = False
            self._window.undoStack().endMacro()
    # end def
    
    def isMacrod(self):
        return self._isMacrod
    # end def
    
    def setMacrod(self):
        self._isMacrod = True
        self._window.undoStack().beginMacro("Group Paint")
        self._window.pathGraphicsView.addToPressList(self)
    # end def



from .abstractpathtool import AbstractPathTool
from cadnano2.views import styles
import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush', 'QFont', 'QPainterPath',
                                       'QPen', 'QPolygonF'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem'])

_bw = styles.PATH_BASE_WIDTH
_pen = QPen(styles.redstroke, 1)
_rect = QRectF(0, 0, _bw, _bw)
_pathArrowLeft = QPainterPath()
_l3poly = QPolygonF()
_l3poly.append(QPointF(_bw, 0))
_l3poly.append(QPointF(0.25 * _bw, 0.5 * _bw))
_l3poly.append(QPointF(_bw, _bw))
_pathArrowLeft.addPolygon(_l3poly)
_pathArrowRight = QPainterPath()
_r3poly = QPolygonF()  # right-hand 3' arr
_r3poly.append(QPointF(0, 0))
_r3poly.append(QPointF(0.75 * _bw, 0.5 * _bw))
_r3poly.append(QPointF(0, _bw))
_pathArrowRight.addPolygon(_r3poly)


class BreakTool(AbstractPathTool):
    """
    docstring for BreakTool
    """
    def __init__(self, controller):
        super(BreakTool, self).__init__(controller)
        self._isTopStrand = True

    def __repr__(self):
        return "breakTool"  # first letter should be lowercase

    def paint(self, painter, option, widget=None):
        super(BreakTool, self).paint(painter, option, widget)
        painter.setPen(_pen)
        if self._isTopStrand:
            painter.drawPath(self._pathArrowRight)
        else:
            painter.drawPath(self._pathArrowLeft)

    def setTopStrand(self, isTop):
        """
        Called in hoverMovePathHelix to set whether breaktool is hovering
        over a top strand (goes 5' to 3' left to right) or bottom strand.
        """
        self._isTopStrand = isTop

    def hoverMove(self, item, event, flag=None):
        """
        flag is for the case where an item in the path also needs to
        implement the hover method
        """
        self.updateLocation(item, item.mapToScene(QPointF(event.position())))
        posScene = item.mapToScene(QPointF(event.position()))
        posItem = item.mapFromScene(posScene)
        self.setTopStrand(self.helixIndex(posItem)[1] == 0)
        newPosition = self.helixPos(posItem)
        if newPosition != None:
            self.setPos(newPosition)



#!/usr/bin/env python
# encoding: utf-8

from math import floor
from cadnano2.controllers.itemcontrollers.virtualhelixitemcontroller import VirtualHelixItemController
from cadnano2.model.enum import StrandType
from .strand.stranditem import StrandItem
from cadnano2.views import styles
from .virtualhelixhandleitem import VirtualHelixHandleItem
import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject', 'Qt', 'QRectF'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QPainterPath',
                                       'QPen',
                                       'QBrush',
                                       'QColor'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem',
                                           'QGraphicsPathItem',
                                           'QGraphicsRectItem'])


_baseWidth = styles.PATH_BASE_WIDTH
# _gridPen = QPen(styles.minorgridstroke, styles.MINOR_GRID_STROKE_WIDTH)
# _gridPen.setCosmetic(True)


class VirtualHelixItem(QGraphicsPathItem):
    """VirtualHelixItem for PathView"""
    findChild = util.findChild  # for debug

    def __init__(self, partItem, modelVirtualHelix, viewroot, activeTool):
        super(VirtualHelixItem, self).__init__(partItem.proxy())
        self._partItem = partItem
        self._modelVirtualHelix = modelVirtualHelix
        self._viewroot = viewroot
        self._activeTool = activeTool
        self._controller = VirtualHelixItemController(self, modelVirtualHelix)

        self._handle = VirtualHelixHandleItem(modelVirtualHelix, partItem, viewroot)
        self._lastStrandSet = None
        self._lastIdx = None
        self._scaffoldBackground = None
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemUsesExtendedStyleOption)
        self.setCacheMode(QGraphicsItem.CacheMode.DeviceCoordinateCache)
        self.setBrush(QBrush(Qt.BrushStyle.NoBrush))

        view = viewroot.scene().views()[0]
        view.levelOfDetailChangedSignal.connect(self.levelOfDetailChangedSlot)
        shouldShowDetails = view.shouldShowDetails()

        pen = QPen(styles.minorgridstroke, styles.MINOR_GRID_STROKE_WIDTH)
        pen.setCosmetic(shouldShowDetails)
        self.setPen(pen)

        self.refreshPath()
        self.setAcceptHoverEvents(True)  # for pathtools
        self.setZValue(styles.ZPATHHELIX)
    # end def

    ### SIGNALS ###

    ### SLOTS ###
    def levelOfDetailChangedSlot(self, boolval):
        """Not connected to the model, only the QGraphicsView"""
        pen = self.pen()
        pen.setCosmetic(boolval)
        self.setPen(pen)
    # end def

    def strandAddedSlot(self, sender, strand):
        """
        Instantiates a StrandItem upon notification that the model has a
        new Strand.  The StrandItem is responsible for creating its own
        controller for communication with the model, and for adding itself to
        its parent (which is *this* VirtualHelixItem, i.e. 'self').
        """
        StrandItem(strand, self, self._viewroot)
    # end def

    def decoratorAddedSlot(self, decorator):
        """
        Instantiates a DecoratorItem upon notification that the model has a
        new Decorator.  The Decorator is responsible for creating its own
        controller for communication with the model, and for adding itself to
        its parent (which is *this* VirtualHelixItem, i.e. 'self').
        """
        pass

    def virtualHelixNumberChangedSlot(self, virtualHelix, number):
        self._handle.setNumber()
    # end def

    def virtualHelixRemovedSlot(self, virtualHelix):
        self._controller.disconnectSignals()
        self._controller = None
        scene = self.scene()
        self._handle.remove()
        scene.removeItem(self)
        self._partItem.removeVirtualHelixItem(self)
        self._partItem = None
        self._modelVirtualHelix = None
        self._activeTool = None
        self._handle = None
    # end def

    ### ACCESSORS ###
    def activeTool(self):
        return self._activeTool

    def coord(self):
        return self._modelVirtualHelix.coord()
    # end def

    def viewroot(self):
        return self._viewroot
    # end def

    def handle(self):
        return self._handle
    # end def

    def part(self):
        return self._partItem.part()
    # end def

    def partItem(self):
        return self._partItem
    # end def

    def number(self):
        if self._modelVirtualHelix is None:
            return ""
        return self._modelVirtualHelix.number()
    # end def

    def virtualHelix(self):
        return self._modelVirtualHelix
    # end def

    def window(self):
        return self._partItem.window()
    # end def

    ### DRAWING METHODS ###
    def isStrandOnTop(self, strand):
        sS = strand.strandSet()
        isEvenParity = self._modelVirtualHelix.isEvenParity()
        return isEvenParity and sS.isScaffold() or\
               not isEvenParity and sS.isStaple()
    # end def

    def isStrandTypeOnTop(self, strandType):
        isEvenParity = self._modelVirtualHelix.isEvenParity()
        return isEvenParity and strandType == StrandType.Scaffold or \
               not isEvenParity and strandType == StrandType.Staple
    # end def

    def upperLeftCornerOfBase(self, idx, strand):
        x = idx * _baseWidth
        y = 0 if self.isStrandOnTop(strand) else _baseWidth
        return x, y
    # end def

    def upperLeftCornerOfBaseType(self, idx, strandType):
        x = idx * _baseWidth
        y = 0 if self.isStrandTypeOnTop(strandType) else _baseWidth
        return x, y
    # end def

    def refreshPath(self):
        """
        Returns a QPainterPath object for the minor grid lines.
        The path also includes a border outline and a midline for
        dividing scaffold and staple bases.
        """
        bw = _baseWidth
        bw2 = 2 * bw
        part = self.part()
        path = QPainterPath()
        subStepSize = part.subStepSize()
        canvasSize = part.maxBaseIdx()+1
        # border
        path.addRect(0, 0, bw * canvasSize, 2 * bw)
        # minor tick marks
        for i in range(canvasSize):
            x = round(bw * i) + .5
            if i % subStepSize == 0:
                path.moveTo(x-.5, 0)
                path.lineTo(x-.5, bw2)
                path.lineTo(x-.25, bw2)
                path.lineTo(x-.25, 0)
                path.lineTo(x, 0)
                path.lineTo(x, bw2)
                path.lineTo(x+.25, bw2)
                path.lineTo(x+.25, 0)
                path.lineTo(x+.5, 0)
                path.lineTo(x+.5, bw2)

                # path.moveTo(x-.5, 0)
                # path.lineTo(x-.5, 2 * bw)
                # path.lineTo(x+.5, 2 * bw)
                # path.lineTo(x+.5, 0)

            else:
                path.moveTo(x, 0)
                path.lineTo(x, 2 * bw)

        # staple-scaffold divider
        path.moveTo(0, bw)
        path.lineTo(bw * canvasSize, bw)

        self.setPath(path)

        if self._modelVirtualHelix.scaffoldIsOnTop():
            scaffoldY = 0
        else:
            scaffoldY = bw
        # if self._scaffoldBackground == None:
        #     highlightr = QGraphicsRectItem(0, scaffoldY, bw * canvasSize, bw, self)
        #     highlightr.setBrush(QBrush(styles.scaffold_bkg_fill))
        #     highlightr.setPen(QPen(Qt.PenStyle.NoPen))
        #     highlightr.setFlag(QGraphicsItem.GraphicsItemFlag.ItemStacksBehindParent)
        #     self._scaffoldBackground = highlightr
        # else:
        #     self._scaffoldBackground.setRect(0, scaffoldY, bw * canvasSize, bw)

    # end def

    def resize(self):
        """Called by part on resize."""
        self.refreshPath()

    ### PUBLIC SUPPORT METHODS ###
    def setActive(self, idx):
        """Makes active the virtual helix associated with this item."""
        self.part().setActiveVirtualHelix(self._modelVirtualHelix, idx)
    # end def

    ### EVENT HANDLERS ###
    def mousePressEvent(self, event):
        """
        Parses a mousePressEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        self.scene().views()[0].addToPressList(self)
        strandSet, idx = self.baseAtPoint(event.pos())
        self.setActive(idx)
        toolMethodName = str(self._activeTool()) + "MousePress"

        ### uncomment for debugging modifier selection
        # strandSet, idx = self.baseAtPoint(event.position())
        # row, col = strandSet.virtualHelix().coord()
        # self._partItem.part().selectPreDecorator([(row,col,idx)])

        if hasattr(self, toolMethodName):
            self._lastStrandSet, self._lastIdx = strandSet, idx
            getattr(self, toolMethodName)(strandSet, idx)
        else:
            event.setAccepted(False)
    # end def

    def mouseMoveEvent(self, event):
        """
        Parses a mouseMoveEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        toolMethodName = str(self._activeTool()) + "MouseMove"
        if hasattr(self, toolMethodName):
            strandSet, idx = self.baseAtPoint(event.pos())
            if self._lastStrandSet != strandSet or self._lastIdx != idx:
                self._lastStrandSet, self._lastIdx = strandSet, idx
                getattr(self, toolMethodName)(strandSet, idx)
        else:
            event.setAccepted(False)
    # end def

    def customMouseRelease(self, event):
        """
        Parses a mouseReleaseEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        toolMethodName = str(self._activeTool()) + "MouseRelease"
        if hasattr(self, toolMethodName):
            getattr(self, toolMethodName)(self._lastStrandSet, self._lastIdx)
        else:
            event.setAccepted(False)
    # end def

    ### COORDINATE UTILITIES ###
    def baseAtPoint(self, pos):
        """
        Returns the (strandType, index) under the location x,y or None.

        It shouldn't be possible to click outside a pathhelix and still call
        this function. However, this sometimes happens if you click exactly
        on the top or bottom edge, resulting in a negative y value.
        """
        x, y = pos.x(), pos.y()
        mVH = self._modelVirtualHelix
        baseIdx = int(floor(x / _baseWidth))
        minBase, maxBase = 0, mVH.part().maxBaseIdx()
        if baseIdx < minBase or baseIdx >= maxBase:
            baseIdx = util.clamp(baseIdx, minBase, maxBase)
        if y < 0:
            y = 0  # HACK: zero out y due to erroneous click
        strandIdx = floor(y * 1. / _baseWidth)
        if strandIdx < 0 or strandIdx > 1:
            strandIdx = int(util.clamp(strandIdx, 0, 1))
        strandSet = mVH.getStrandSetByIdx(strandIdx)
        return (strandSet, baseIdx)
    # end def

    def keyPanDeltaX(self):
        """How far a single press of the left or right arrow key should move
        the scene (in scene space)"""
        dx = self._partItem.part().stepSize() * _baseWidth
        return self.mapToScene(QRectF(0, 0, dx, 1)).boundingRect().width()
    # end def

    def hoverLeaveEvent(self, event):
        self._partItem.updateStatusBar("")
    # end def

    def hoverMoveEvent(self, event):
        """
        Parses a mouseMoveEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        baseIdx = int(floor(event.pos().x() / _baseWidth))
        loc = "%d[%d]" % (self.number(), baseIdx)
        self._partItem.updateStatusBar(loc)

        activeTool = self._activeTool()
        toolMethodName = str(activeTool) + "HoverMove"
        if hasattr(self, toolMethodName):
            strandType, idxX, idxY = activeTool.baseAtPoint(self, event.pos())
            getattr(self, toolMethodName)(strandType, idxX, idxY)
    # end def

    ### TOOL METHODS ###
    def pencilToolMousePress(self, strandSet, idx):
        """strand.getDragBounds"""
        # print "%s: %s[%s]" % (util.methodName(), strandSet, idx)
        activeTool = self._activeTool()
        if not activeTool.isDrawingStrand():
            activeTool.initStrandItemFromVHI(self, strandSet, idx)
            activeTool.setIsDrawingStrand(True)
    # end def

    def pencilToolMouseMove(self, strandSet, idx):
        """strand.getDragBounds"""
        # print "%s: %s[%s]" % (util.methodName(), strandSet, idx)
        activeTool = self._activeTool()
        if activeTool.isDrawingStrand():
            activeTool.updateStrandItemFromVHI(self, strandSet, idx)
    # end def

    def pencilToolMouseRelease(self, strandSet, idx):
        """strand.getDragBounds"""
        # print "%s: %s[%s]" % (util.methodName(), strandSet, idx)
        activeTool = self._activeTool()
        if activeTool.isDrawingStrand():
            activeTool.setIsDrawingStrand(False)
            activeTool.attemptToCreateStrand(self, strandSet, idx)
    # end def

    def pencilToolHoverMove(self, strandType, idxX, idxY):
        """Pencil the strand is possible."""
        partItem = self.partItem()
        activeTool = self._activeTool()
        if not activeTool.isFloatingXoverBegin():
            tempXover = activeTool.floatingXover()
            tempXover.updateFloatingFromVHI(self, strandType, idxX, idxY)
    # end def



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
HalfCylinderHelixNode.py
Created by Simon Breslav on 2011-08-04
A Maya Node for creating Half-Cylinder Helix Shape
"""

import sys
import maya.OpenMaya as OpenMaya
import maya.OpenMayaMPx as OpenMayaMPx
import math

nodeName = "spHalfCylinderHelixNode"
id = OpenMaya.MTypeId(0x00117701)


class HalfCylinderHelixNode(OpenMayaMPx.MPxNode):
    """A Maya Node for creating Half-Cylinder Helix 3D Shape"""

    outputMesh = OpenMaya.MObject()
    startBaseAttr = OpenMaya.MObject()
    endBaseAttr = OpenMaya.MObject()
    startBaseFloatAttr = OpenMaya.MObject()
    endBaseFloatAttr = OpenMaya.MObject()
    totalBasesAttr = OpenMaya.MObject()
    start3DPosAttr = OpenMaya.MObject()
    end3DPosAttr = OpenMaya.MObject()
    parityAttr = OpenMaya.MObject()
    rotationOffsetAttr = OpenMaya.MObject()
    decRotOffsetAttr = OpenMaya.MObject()
    radiusAttr = OpenMaya.MObject()
    strandTypeAttr = OpenMaya.MObject()
    rotationAttr = OpenMaya.MObject()
    riseAttr = OpenMaya.MObject()
    edgesPerBaseAttr = OpenMaya.MObject()

    def __init__(self):
        """Initiate the Node and private variable"""
        OpenMayaMPx.MPxNode.__init__(self)
        self.start3DPos = OpenMaya.MFloatPoint()
        self.end3DPos = OpenMaya.MFloatPoint()

    def compute(self, plug, data):
        """
        Main compute method that is called each time the shape needs
        to be recomputed, such as change of some attributes, these recompute
        dependencies are created by HalfCylinderHelixNode.attributeAffects
        calls in nodeInitialize() function
        """
        try:
            fnMeshData = OpenMaya.MFnMeshData()
            outMeshDataObj = fnMeshData.create()
            endData = data.inputValue(HalfCylinderHelixNode.endBaseAttr)
            startData = data.inputValue(HalfCylinderHelixNode.startBaseAttr)
            startVal = startData.asInt()
            endVal = endData.asInt()

            rotationData = data.inputValue(HalfCylinderHelixNode.rotationAttr)
            spacingData = data.inputValue(HalfCylinderHelixNode.riseAttr)
            edgesPerBaseData = \
                data.inputValue(HalfCylinderHelixNode.edgesPerBaseAttr)
            radiusData = data.inputValue(HalfCylinderHelixNode.radiusAttr)
            totalNumBasesData = \
                data.inputValue(HalfCylinderHelixNode.totalBasesAttr)
            totalNumBases = totalNumBasesData.asInt()
            parityData = data.inputValue(HalfCylinderHelixNode.parityAttr)
            parity = parityData.asInt()
            rotationOffsetData = \
                data.inputValue(HalfCylinderHelixNode.rotationOffsetAttr)
            rotationOffset = rotationOffsetData.asDouble()
            strandTypeData = \
                data.inputValue(HalfCylinderHelixNode.strandTypeAttr)

            self.createMesh(startVal, endVal, totalNumBases,
                            radiusData.asDouble(), rotationData.asDouble(),
                            rotationOffset,
                            spacingData.asDouble(),
                            edgesPerBaseData.asInt(),
                            parity,
                            strandTypeData.asInt(),
                            outMeshDataObj)

            startPosHandle = \
                data.outputValue(HalfCylinderHelixNode.start3DPosAttr)
            startPosHandle.set3Float(self.start3DPos[0],
                                      self.start3DPos[1],
                                      self.start3DPos[2])
            endPosHandle = \
                data.outputValue(HalfCylinderHelixNode.end3DPosAttr)
            endPosHandle.set3Float(self.end3DPos[0],
                                    self.end3DPos[1],
                                    self.end3DPos[2])

            handle = data.outputValue(HalfCylinderHelixNode.outputMesh)
            handle.setMObject(outMeshDataObj)
            handle.setClean()
            data.setClean(plug)
        except:
            print("Error in %s\n" % nodeName)
            raise

    def createMesh(self, startVal, endVal, totalNumBases, radius,
                   rotationAttr, rotationOffset, riseAttr, edgesPerBase,
                   parity, strandType, outData):
        """
        This is the core function of the class that created the MFnMesh of
        the Half-Cylinder Helical Shape. The process can be though of as
        if we are creating a regular cylinder, so first create the starting
        endcap, then create the middle part, and last the closing endcap.
        When creating the middle part, we add additional rotation to the
        verts to create the overall helical shape.
        """
        # XXX [SB] are start and end are inverted right now?
        middleBase = totalNumBases / 2
        end = (middleBase - startVal)
        start = (endVal - middleBase)
        baseCount = start + end + 1
        numVerticesEnds = 20
        numMiddleSections = int(baseCount * edgesPerBase) - 1
        rise = riseAttr / edgesPerBase
        start_pos = -start * (riseAttr)
        numVerticesTotal = (numVerticesEnds * 2) + \
                            numMiddleSections * numVerticesEnds
        rot_ang = -rotationAttr / edgesPerBase
        gap = rise / 6.0

        numFacesEnds = numVerticesEnds - 2
        numFacesTotal = (numVerticesEnds - 2) * 2 + \
                        (numVerticesEnds) * (numMiddleSections + 1)
        numFaceConnects = ((numVerticesEnds - 2) * 2 * 3) + \
                           (numVerticesEnds * (numMiddleSections + 1) * 4)

        vtx = []
        rotation = 0
        starting_rotation = endVal * rotationAttr + \
                            (math.pi * (not parity)) + rotationOffset + \
                            (math.pi * strandType)

        # Create Endpoint verts
        vtx.append(OpenMaya.MFloatPoint(0.0, start_pos + gap, 0.0))
        self.end3DPos = OpenMaya.MFloatPoint(0.0, start_pos + gap, 0.0)
        #print start_pos + gap
        for i in range(1, numVerticesEnds):
            val = i * (180 / (numFacesEnds))
            rad = (val * math.pi) / 180
            vtx.append(OpenMaya.MFloatPoint(radius * \
                                math.cos(starting_rotation + rad),
                                start_pos + gap,
                                radius * math.sin(starting_rotation + rad)))
        # Create Middle verts
        for i in range(0, numMiddleSections):
            rotation = rot_ang * (i + 1)
            pos = rise * (i + 1)
            vtx.append(OpenMaya.MFloatPoint(0.0, start_pos + pos, 0.0))
            for i in range(1, numVerticesEnds):
                val = i * (180 / (numFacesEnds))
                rad = (val * math.pi) / 180
                #print "vrt %d, %f %f" % (i, val, rad)
                x = radius * math.cos(starting_rotation + rad + rotation)
                y = radius * math.sin(starting_rotation + rad + rotation)
                vtx.append(OpenMaya.MFloatPoint(x, start_pos + pos, y))

        # Create EndPiece verts
        vtx.append(OpenMaya.MFloatPoint(0.0,
                            start_pos + (numMiddleSections + 1) * rise - gap,
                            0.0))
        self.start3DPos = OpenMaya.MFloatPoint(0.0,
                            start_pos + (numMiddleSections + 1) * rise - gap,
                             0.0)
        for i in range(1, numVerticesEnds):
            rotation = rot_ang * (1 + numMiddleSections)
            val = i * (180 / (numFacesEnds))
            rad = (val * math.pi) / 180
            #print "vrt %d, %f %f" % (i, val, rad)
            x = radius * math.cos(starting_rotation + rad + rotation)
            y = radius * math.sin(starting_rotation + rad + rotation)
            vtx.append(OpenMaya.MFloatPoint(x,
                            start_pos + (numMiddleSections + 1) * rise - gap,
                            y))
        #print "vtx length %d"  % len(vtx)
        points = OpenMaya.MFloatPointArray()
        points.setLength(numVerticesTotal)

        for i in range(0, numVerticesTotal):
            points.set(vtx[i], i)

        faceConnects = OpenMaya.MIntArray()
        faceConnects.setLength(numFaceConnects)
        #print "numFaceConnects %d" % numFaceConnects
        offset = 1
        count = 0

        # Front End Piece Faces
        # in a fan fashion, we build a list of faces, each face is 3 verts,
        # with 0 being the verst vet, so we get faces such as (0, 1, 2),
        # the next face would be (0, 2, 3), and next (0, 3, 4), etc.
        for i in range(0, numFacesEnds):
            faceConnects.set(0, count)
            count += 1
            faceConnects.set(offset, count)
            offset += 1
            count += 1
            faceConnects.set(offset, count)
            count += 1

        # Middle Piece Faces
        # 
        for k in range(0, numMiddleSections + 1):
            for i in range(0, numVerticesEnds):
                if i < numVerticesEnds - 1:
                    v1 = (k * numVerticesEnds) + i
                    v2 = (k * numVerticesEnds) + 1 + i
                    v3 = ((k + 1) * numVerticesEnds) + i
                    v4 = ((k + 1) * numVerticesEnds) + 1 + i
                else:
                    v1 = (k * numVerticesEnds) + i
                    v2 = (k * numVerticesEnds)
                    v3 = ((k + 1) * numVerticesEnds) + i
                    v4 = ((k + 1) * numVerticesEnds)
                faceConnects.set(v1, count)
                count += 1
                faceConnects.set(v2, count)
                count += 1
                faceConnects.set(v4, count)
                count += 1
                faceConnects.set(v3, count)
                count += 1

        # Back End Piece Faces, like the front one, we do it in a fan order
        offset = numVerticesTotal - numVerticesEnds + 1
        for i in range(numFacesTotal - numFacesEnds, numFacesTotal):
            #print numVerticesTotal-numVerticesEnds
            #print offset
            faceConnects.set(numVerticesTotal - numVerticesEnds, count)
            count += 1
            faceConnects.set(offset, count)
            offset += 1
            count += 1
            #print offset
            faceConnects.set(offset, count)
            count += 1
        #print "face connect ended at %d" % count
        faceCounts = OpenMaya.MIntArray()
        faceCounts.setLength(numFacesTotal)
        count = 0
        # Set Face Counds, the faces on the ends are made from 3 verts
        # and faces in the middle are made from 4
        for i in range(0, numFacesEnds):
            #print count
            faceCounts.set(3, count)
            count += 1
        for i in range(numFacesEnds, numFacesTotal - numFacesEnds):
            #print count
            faceCounts.set(4, count)
            count += 1
        for i in range(numFacesTotal - numFacesEnds, numFacesTotal):
            #print count
            faceCounts.set(3, count)
            count += 1
        #print "face count ended at %d" % count
        meshFS = OpenMaya.MFnMesh()
        meshFS.create(numVerticesTotal, numFacesTotal,
                      points, faceCounts, faceConnects, outData)


def nodeCreator():
    """Creates the Maya Node"""
    return OpenMayaMPx.asMPxPtr(HalfCylinderHelixNode())


def nodeInitialize():
    """
    This initialize function adds all the attributes and creates relationships
    between them, so that as you change attributes related attributes are
    recomputed through the compute function of the node
    """
    #unitAttr = OpenMaya.MFnUnitAttribute()
    typedAttr = OpenMaya.MFnTypedAttribute()
    HalfCylinderHelixNode.outputMesh = typedAttr.create("outputMesh",
                                                "out",
                                                OpenMaya.MFnData.kMesh)

    nAttr = OpenMaya.MFnNumericAttribute()

    HalfCylinderHelixNode.startBaseAttr = nAttr.create('startBase',
                                    'stb',
                                    OpenMaya.MFnNumericData.kInt,
                                    0)
    nAttr.setStorable(True)
    HalfCylinderHelixNode.endBaseAttr = nAttr.create('endBase',
                                    'sab',
                                    OpenMaya.MFnNumericData.kInt,
                                    0)
    nAttr.setStorable(True)

    HalfCylinderHelixNode.startBaseFloatAttr = nAttr.create('startBaseFloat',
                                    'stbf',
                                    OpenMaya.MFnNumericData.kDouble,
                                    0)
    nAttr.setStorable(True)
    nAttr.setWritable(True)

    HalfCylinderHelixNode.endBaseFloatAttr = nAttr.create('endBaseFloat',
                                    'sabf',
                                    OpenMaya.MFnNumericData.kDouble,
                                    0)
    nAttr.setStorable(True)
    nAttr.setWritable(True)

    HalfCylinderHelixNode.totalBasesAttr = nAttr.create('totalBases',
                                    'tb',
                                    OpenMaya.MFnNumericData.kInt,
                                    0)
    nAttr.setStorable(True)
    nAttr.setWritable(True)

    HalfCylinderHelixNode.riseAttr = nAttr.create('rise',
                                    'r',
                                    OpenMaya.MFnNumericData.kDouble,
                                    0.34)
    nAttr.setMin(0.01)
    nAttr.setMax(1.0)
    nAttr.setStorable(True)
    HalfCylinderHelixNode.edgesPerBaseAttr = nAttr.create('edgesPerBase',
                                    'spb',
                                    OpenMaya.MFnNumericData.kInt,
                                    2)
    nAttr.setMin(1)
    nAttr.setMax(20)
    nAttr.setStorable(True)

    HalfCylinderHelixNode.radiusAttr = nAttr.create('radius',
                                    'rad',
                                    OpenMaya.MFnNumericData.kDouble,
                                    1.125)
    nAttr.setMin(1.0)
    nAttr.setMax(5.0)
    nAttr.setStorable(True)

    HalfCylinderHelixNode.parityAttr = nAttr.create('parity',
                                            'pa',
                                            OpenMaya.MFnNumericData.kInt,
                                            0)
    nAttr.setStorable(True)

    HalfCylinderHelixNode.strandTypeAttr = nAttr.create('strandType',
                                            'st',
                                            OpenMaya.MFnNumericData.kInt,
                                            0)
    nAttr.setStorable(True)

    HalfCylinderHelixNode.start3DPosAttr = nAttr.create('startPos',
                                            'sp',
                                            OpenMaya.MFnNumericData.k3Float,
                                            0.0)
    nAttr.setStorable(False)
    HalfCylinderHelixNode.end3DPosAttr = nAttr.create('endPos',
                                            'ep',
                                            OpenMaya.MFnNumericData.k3Float,
                                            0.0)
    nAttr.setStorable(False)

    unitFn = OpenMaya.MFnUnitAttribute()
    HalfCylinderHelixNode.rotationAttr = unitFn.create('rotation',
                                    'rot',
                                    OpenMaya.MFnUnitAttribute.kAngle,
                                    34.286 * math.pi / 180)
    unitFn.setMin(0.0)
    unitFn.setMax(math.pi)
    unitFn.setStorable(True)
    HalfCylinderHelixNode.rotationOffsetAttr = unitFn.create('rotationOffset',
                                    'rotoff',
                                    OpenMaya.MFnUnitAttribute.kAngle,
                                    math.pi / 6)
    unitFn.setMin(-2 * math.pi)
    unitFn.setMax(2 * math.pi)
    unitFn.setStorable(True)

    HalfCylinderHelixNode.decRotOffsetAttr = unitFn.create(
                                    'decoratorRotOffset',
                                    'decrotoff',
                                    OpenMaya.MFnUnitAttribute.kAngle,
                                    math.pi / 9)
    unitFn.setMin(-2 * math.pi)
    unitFn.setMax(2 * math.pi)
    unitFn.setStorable(True)

    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.startBaseAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.endBaseAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.startBaseFloatAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.endBaseFloatAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.totalBasesAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.rotationAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.rotationOffsetAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.decRotOffsetAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.riseAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.edgesPerBaseAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.start3DPosAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.end3DPosAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.radiusAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.parityAttr)
    HalfCylinderHelixNode.addAttribute(HalfCylinderHelixNode.strandTypeAttr)

    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.endBaseAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.startBaseAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.endBaseFloatAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.startBaseFloatAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.totalBasesAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.rotationAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.riseAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.edgesPerBaseAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.radiusAttr,
        HalfCylinderHelixNode.outputMesh)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.rotationOffsetAttr,
        HalfCylinderHelixNode.outputMesh)

    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.endBaseAttr,
        HalfCylinderHelixNode.start3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.startBaseAttr,
        HalfCylinderHelixNode.start3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.endBaseFloatAttr,
        HalfCylinderHelixNode.start3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.startBaseFloatAttr,
        HalfCylinderHelixNode.start3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.totalBasesAttr,
        HalfCylinderHelixNode.start3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.riseAttr,
        HalfCylinderHelixNode.start3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.edgesPerBaseAttr,
        HalfCylinderHelixNode.start3DPosAttr)

    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.endBaseAttr,
        HalfCylinderHelixNode.end3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.startBaseAttr,
        HalfCylinderHelixNode.end3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.endBaseFloatAttr,
        HalfCylinderHelixNode.end3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.startBaseFloatAttr,
        HalfCylinderHelixNode.end3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.totalBasesAttr,
        HalfCylinderHelixNode.end3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.riseAttr,
        HalfCylinderHelixNode.end3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.edgesPerBaseAttr,
        HalfCylinderHelixNode.start3DPosAttr)

    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.outputMesh,
        HalfCylinderHelixNode.start3DPosAttr)
    HalfCylinderHelixNode.attributeAffects(
        HalfCylinderHelixNode.outputMesh,
        HalfCylinderHelixNode.end3DPosAttr)


def initializePlugin(obj):
    """Initialize the Plugin"""
    plugin = OpenMayaMPx.MFnPlugin(obj)
    try:
        plugin.registerNode(nodeName, id, nodeCreator, nodeInitialize)
    except:
        sys.stderr.write("Failed to register node %s\n" % nodeName)
        raise


def uninitializePlugin(obj):
    """Uninitialize the Plugin"""
    plugin = OpenMayaMPx.MFnPlugin(obj)
    try:
        plugin.deregisterNode(id)
    except:
        sys.stderr.write("Failed to deregister node %s\n" % nodeName)
        raise



from cadnano2.views import styles
from cadnano2.model.virtualhelix import VirtualHelix
from cadnano2.model.enum import Parity, StrandType
from cadnano2.controllers.itemcontrollers.virtualhelixitemcontroller import VirtualHelixItemController

import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['Qt', 'QEvent', 'QRectF', 'QPointF'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QPen',
                                       'QPainterPath',
                                       'QPolygonF',
                                       'QColor'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsEllipseItem',
                                           'QGraphicsItem',
                                           'QGraphicsLineItem',
                                           'QGraphicsSimpleTextItem',
                                           ])


class VirtualHelixItem(QGraphicsEllipseItem):
    """
    The VirtualHelixItem is an individual circle that gets drawn in the SliceView
    as a child of the PartItem. Taken as a group, many SliceHelix
    instances make up the crossection of the DNAPart. Clicking on a SliceHelix
    adds a VirtualHelix to the DNAPart. The SliceHelix then changes appearence
    and paints its corresponding VirtualHelix number.
    """
    # set up default, hover, and active drawing styles
    _useBrush = QBrush(styles.orangefill)
    _usePen = QPen(styles.orangestroke, styles.SLICE_HELIX_STROKE_WIDTH)
    _radius = styles.SLICE_HELIX_RADIUS
    _outOfSlicePen = QPen(styles.lightorangestroke,
                          styles.SLICE_HELIX_STROKE_WIDTH)
    _outOfSliceBrush = QBrush(styles.lightorangefill)
    _rect = QRectF(0, 0, 2 * _radius, 2 * _radius)
    _font = styles.SLICE_NUM_FONT
    _ZVALUE = styles.ZSLICEHELIX+3

    def __init__(self, modelVirtualHelix, emptyHelixItem):
        """
        emptyHelixItem is a EmptyHelixItem that will act as a QGraphicsItem parent
        """
        super(VirtualHelixItem, self).__init__(parent=emptyHelixItem)
        self._virtualHelix = modelVirtualHelix
        self._emptyHelixItem = emptyHelixItem
        self.hide()
        # drawing related

        self.isHovered = False
        self.setAcceptHoverEvents(True)
        # self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setZValue(self._ZVALUE)
        self.lastMousePressAddedBases = False

        self.setBrush(self._outOfSliceBrush)
        self.setPen(self._outOfSlicePen)
        self.setRect(self._rect)

        # handle the label specific stuff
        self._label = self.createLabel()
        self.setNumber()
        self.createArrow()

        self._controller = VirtualHelixItemController(self, modelVirtualHelix)

        self.show()
    # end def

    ### SIGNALS ###

    ### SLOTS ###
    def virtualHelixNumberChangedSlot(self, virtualHelix):
        """
        receives a signal containing a virtualHelix and the oldNumber 
        as a safety check
        """
        self.setNumber()
    # end def

    def virtualHelixRemovedSlot(self, virtualHelix):
        self._controller.disconnectSignals()
        self._controller = None
        self._emptyHelixItem.setNotHovered()
        self._virtualHelix = None
        self._emptyHelixItem = None
        self.scene().removeItem(self._label)
        self._label = None
        self.scene().removeItem(self)
    # end def

    def strandAddedSlot(self, sender, strand):
        pass
    # end def

    ###

    def createLabel(self):
        label = QGraphicsSimpleTextItem("%d" % self._virtualHelix.number())
        label.setFont(self._font)
        label.setZValue(self._ZVALUE)
        label.setParentItem(self)
        return label
    # end def

    def createArrow(self):
        rad = self._radius
        pen = QPen()
        pen.setWidth(3)
        color = QColor(Qt.GlobalColor.blue)
        color.setAlphaF(0.25)
        pen.setBrush(color)
        if self._virtualHelix.isEvenParity():
            arrow = QGraphicsLineItem(rad, rad, 2*rad, rad, self)
        else:
            arrow = QGraphicsLineItem(0, rad, rad, rad, self)
        arrow.setTransformOriginPoint(rad, rad)
        arrow.setZValue(400)
        arrow.setPen(pen)
        self.arrow = arrow
        self.arrow.hide()
    # end def

    def updateArrow(self, idx):
        part = self.part()
        tpb = part._twistPerBase
        angle = idx*tpb
        self.arrow.setRotation(angle + part._twistOffset)
    # end def

    def setNumber(self):
        """docstring for setNumber"""
        vh = self._virtualHelix
        num = vh.number()
        label = self._label
        radius = self._radius

        if num != None:
            label.setText("%d" % num)
        else:
            return

        y_val = radius / 3
        if num < 10:
            label.setPos(radius / 1.5, y_val)
        elif num < 100:
            label.setPos(radius / 3, y_val)
        else: # _number >= 100
            label.setPos(0, y_val)
        bRect = label.boundingRect()
        posx = bRect.width()/2
        posy = bRect.height()/2
        label.setPos(radius-posx, radius-posy)
    # end def

    def part(self):
        return self._emptyHelixItem.part()

    def virtualHelix(self):
        return self._virtualHelix
    # end def

    def number(self):
        return self.virtualHelix().number()

    def setActiveSliceView(self, isActiveNow, idx):
        if isActiveNow:
            self.setPen(self._usePen)
            self.setBrush(self._useBrush)
            self.updateArrow(idx)
            self.arrow.show()
        else:
            self.setPen(self._outOfSlicePen)
            self.setBrush(self._outOfSliceBrush)
            self.arrow.hide()
    # end def

    ############################ User Interaction ############################
    def sceneEvent(self, event):
        """Included for unit testing in order to grab events that are sent
        via QGraphicsScene.sendEvent()."""
        # if self._parent.sliceController.testRecorder:
        #     coord = (self._row, self._col)
        #     self._parent.sliceController.testRecorder.sliceSceneEvent(event, coord)
        if event.type() == QEvent.Type.MouseButtonPress:
            self.mousePressEvent(event)
            return True
        elif event.type() == QEvent.Type.MouseButtonRelease:
            self.mouseReleaseEvent(event)
            return True
        elif event.type() == QEvent.Type.MouseMove:
            self.mouseMoveEvent(event)
            return True
        QGraphicsItem.sceneEvent(self, event)
        return False

    def hoverEnterEvent(self, event):
        """
        If the selection is configured to always select
        everything, we don't draw a focus ring around everything,
        instead we only draw a focus ring around the hovered obj.
        """
        # if self.selectAllBehavior():
        #     self.setSelected(True)
        # forward the event to the emptyHelixItem as well
        self._emptyHelixItem.hoverEnterEvent(event)
    # end def

    def hoverLeaveEvent(self, event):
        # if self.selectAllBehavior():
        #     self.setSelected(False)
        self._emptyHelixItem.hoverEnterEvent(event)
    # end def

    # def mousePressEvent(self, event):
    #     action = self.decideAction(event.modifiers())
    #     action(self)
    #     self.dragSessionAction = action
    # 
    # def mouseMoveEvent(self, event):
    #     parent = self._helixItem
    #     posInParent = parent.mapFromItem(self, QPointF(event.position()))
    #     # Qt doesn't have any way to ask for graphicsitem(s) at a
    #     # particular position but it *can* do intersections, so we
    #     # just use those instead
    #     parent.probe.setPos(posInParent)
    #     for ci in parent.probe.collidingItems():
    #         if isinstance(ci, SliceHelix):
    #             self.dragSessionAction(ci)
    # # end def

    # def mouseReleaseEvent(self, event):
    #     self.part().needsFittingToView.emit()

    # def decideAction(self, modifiers):
    #     """ On mouse press, an action (add scaffold at the active slice, add
    #     segment at the active slice, or create virtualhelix if missing) is
    #     decided upon and will be applied to all other slices happened across by
    #     mouseMoveEvent. The action is returned from this method in the form of a
    #     callable function."""
    #     vh = self.virtualHelix()
    #     if vh == None: return SliceHelix.addVHIfMissing
    #     idx = self.part().activeSlice()
    #     if modifiers & Qt.KeyboardModifier.ShiftModifier:
    #         if vh.stap().get(idx) == None:
    #             return SliceHelix.addStapAtActiveSliceIfMissing
    #         else:
    #             return SliceHelix.nop
    #     if vh.scaf().get(idx) == None:
    #         return SliceHelix.addScafAtActiveSliceIfMissing
    #     return SliceHelix.nop
    # 
    # def nop(self):
    #     pass



from cadnano2.controllers.viewrootcontroller import ViewRootController
from .partitem import PartItem
from .pathselection import SelectionItemGroup
from .pathselection import VirtualHelixHandleSelectionBox
from .pathselection import EndpointHandleSelectionBox
import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsRectItem'])


class PathRootItem(QGraphicsRectItem):
    """
    PathRootItem is the root item in the PathView. It gets added directly
    to the pathscene by DocumentWindow. It receives two signals
    (partAddedSignal and documentSelectedPartChangedSignal)
    via its ViewRootController.

    PathRootItem must instantiate its own controller to receive signals
    from the model.
    """
    findChild = util.findChild  # for debug

    def __init__(self, rect, parent, window, document):
        super(PathRootItem, self).__init__(rect, parent)
        self._window = window
        self._document = document
        self._controller = ViewRootController(self, document)
        self._modelPart = None
        self._partItemForPart = {}  # Maps Part -> PartItem
        self._selectionFilterDict = {}
        self._initSelections()
    # end def

    ### SIGNALS ###

    ### SLOTS ###
    def partItems(self):
        return list(self._partItemForPart.values())

    def partItemForPart(self, part):
        return self._partItemForPart[part]

    def partAddedSlot(self, sender, modelPart):
        """
        Receives notification from the model that a part has been added.
        The Pathview doesn't need to do anything on part addition, since
        the Sliceview handles setting up the appropriate lattice.
        """
        # print "PathRootItem partAddedSlot", modelPart
        self._modelPart = modelPart
        win = self._window
        partItem = PartItem(modelPart,\
                            viewroot=self, \
                            activeTool=win.pathToolManager.activeTool,\
                            parent=self)
        self._partItemForPart[modelPart] = partItem
        win.pathToolManager.setActivePart(partItem)
        self.setModifyState(win.actionModify.isChecked())
    # end def

    def selectedChangedSlot(self, itemDict):
        """Given a newly selected modelPart, update the scene to indicate
        that modelPart is selected and the previously selected part is
        deselected."""
        for item, value in itemDict:
            item.selectionProcess(value)
    # end def

    def clearSelectionsSlot(self, doc):
        self._vhiHSelectionGroup.resetSelection()
        self._strandItemSelectionGroup.resetSelection()
        self.scene().views()[0].clearSelectionLockAndCallbacks()
    # end def

    def selectionFilterChangedSlot(self, filterNameList):
        self._vhiHSelectionGroup.clearSelection(False)
        self._strandItemSelectionGroup.clearSelection(False)
        self.clearSelectionFilterDict()
        for filterName in filterNameList:
            self.addToSelectionFilterDict(filterName)
    # end def

    def resetRootItemSlot(self, doc):
        self._vhiHSelectionGroup.resetSelection()
        self._strandItemSelectionGroup.resetSelection()
        self.scene().views()[0].clearGraphicsView()
    # end def

    ### ACCESSORS ###
    def sliceToolManager(self):
        """
        Used for getting access to button signals that need to be connected
        to item slots.
        """
        return self._window.sliceToolManager
    # end def

    def window(self):
        return self._window
    # end def

    def document(self):
        return self._document
    # end def

    def _initSelections(self):
        """Initialize anything related to multiple selection."""
        bType = VirtualHelixHandleSelectionBox
        self._vhiHSelectionGroup = SelectionItemGroup(boxtype=bType,\
                                                      constraint='y',\
                                                      parent=self)
        bType = EndpointHandleSelectionBox
        self._strandItemSelectionGroup = SelectionItemGroup(boxtype=bType,\
                                                      constraint='x',\
                                                      parent=self)
    # end def

    ### PUBLIC METHODS ###
    def getSelectedPartOrderedVHList(self):
        """Used for encoding."""
        try:
            selectedPart = self._document.selectedPart()
            return self._partItemForPart[selectedPart].getOrderedVirtualHelixList()
        except KeyError:
            return None
    # end def

    def removePartItem(self, partItem):
        for k in self._partItemForPart.keys():
            if k==partItem:
                del self._partItemForPart[k]
                return
    # end def

    def resetDocumentAndController(self, document):
        """docstring for resetDocumentAndController"""
        self._document = document
        self._controller = ViewRootController(self, document)
    # end def

    def setModifyState(self, bool):
        """docstring for setModifyState"""
        for partItem in self._partItemForPart.values():
            partItem.setModifyState(bool)
    # end def

    def selectionFilterDict(self):
        return self._selectionFilterDict
    # end def

    def addToSelectionFilterDict(self, filterName):
        self._selectionFilterDict[filterName] = True
    # end def

    def removeFromSelectionFilterDict(self, filterName):
        del self._selectionFilterDict[filterName]
    # end def

    def clearSelectionFilterDict(self):
        self._selectionFilterDict = {}
    # end def

    def vhiHandleSelectionGroup(self):
        return self._vhiHSelectionGroup
    # end def

    def strandItemSelectionGroup(self):
        return self._strandItemSelectionGroup
    # end def

    def selectionLock(self):
        return self.scene().views()[0].selectionLock()
    # end def

    def setSelectionLock(self, locker):
        self.scene().views()[0].setSelectionLock(locker)
    # end def

    def clearStrandSelections(self):
        self._strandItemSelectionGroup.clearSelection(False)
    # end def
# end class



import re
from .abstractpathtool import AbstractPathTool
from cadnano2.data.dnasequences import sequences
from cadnano2.ui.dialogs.ui_addseq import Ui_AddSeqDialog
from cadnano2.views import styles
import cadnano2.util as util

util.qtWrapImport('QtCore', globals(), ['Qt', 'QObject', 'QPointF',
                                        'QRegularExpression', 'QSignalMapper',
                                        'pyqtSignal', 'pyqtSlot'])
util.qtWrapImport('QtGui', globals(), ['QBrush', 'QColor', 'QFont', 'QPen',
                                       'QSyntaxHighlighter',
                                       'QTextCharFormat'])
util.qtWrapImport('QtWidgets', globals(), ['QDialog', 'QDialogButtonBox',
                                           'QRadioButton', ])

dnapattern = QRegularExpression("([^ACGTacgt]+)")


class DNAHighlighter(QSyntaxHighlighter):
    def __init__(self, parent):
        QSyntaxHighlighter.__init__(self, parent)
        self.parent = parent
        self.format = QTextCharFormat()
        self.format.setBackground(QBrush(styles.INVALID_DNA_COLOR))
        if styles.UNDERLINE_INVALID_DNA:
            self.format.setFontUnderline(True)
            self.format.setUnderlineColor(styles.INVALID_DNA_COLOR)

    def highlightBlock(self, text):
        # This needs to be rewritten for QRegularExpression
        # It should highlight all bad blocks, not just the first one
        matchIter = dnapattern.globalMatch(text)


        while matchIter.hasNext():
            match = matchIter.next()
            index = match.capturedStart()
            length = match.capturedLength()
            # try:
            #     index = text.index(str(dnapattern), index + length)
            # except ValueError:
            #     index = -1
            self.setFormat(index, length, self.format)
        self.setCurrentBlockState(0)


class AddSeqTool(AbstractPathTool):
    def __init__(self, controller, parent=None):
        AbstractPathTool.__init__(self, controller, parent)
        self.dialog = QDialog()
        self.buttons = []
        self.seqBox = None
        self.chosenStandardSequence = None  # state for tab switching
        self.customSequenceIsValid = False  # state for tab switching
        self.useCustomSequence = False  # for applying sequence
        self.validatedSequenceToApply = None
        self.initDialog()

    def __repr__(self):
        return "addSeqTool"  # first letter should be lowercase

    def initDialog(self):
        """
        1. Create buttons according to available scaffold sequences and
        add them to the dialog.
        2. Map the clicked signal of those buttons to keep track of what
        sequence gets selected.
        3. Watch the tabWidget change signal to determine whether a
        standard or custom sequence should be applied.
        """
        uiDlg = Ui_AddSeqDialog()
        uiDlg.setupUi(self.dialog)
        self.signalMapper = QSignalMapper(self)
        # set up the radio buttons
        for i, name in enumerate(sorted(sequences.keys())):
            radioButton = QRadioButton(uiDlg.groupBox)
            radioButton.setObjectName(name + "Button")
            radioButton.setText(name)
            self.buttons.append(radioButton)
            uiDlg.verticalLayout.addWidget(radioButton)
            self.signalMapper.setMapping(radioButton, i)
            radioButton.clicked.connect(self.signalMapper.map)
        self.signalMapper.mappedInt.connect(self.standardSequenceChangedSlot)
        uiDlg.tabWidget.currentChanged.connect(self.tabWidgetChangedSlot)
        # disable apply until valid option or custom sequence is chosen
        self.applyButton = uiDlg.customButtonBox.button(QDialogButtonBox.StandardButton.Apply)
        self.applyButton.setEnabled(False)
        # watch sequence textedit box to validate custom sequences
        self.seqBox = uiDlg.seqTextEdit
        self.seqBox.textChanged.connect(self.validateCustomSequence)
        self.highlighter = DNAHighlighter(self.seqBox)
        # finally, pre-click the M13mp18 radio button
        self.buttons[0].click()
        buttons = self.buttons

        self.dialog.setFocusProxy(uiDlg.groupBox)
        self.dialog.setFocusPolicy(Qt.FocusPolicy.TabFocus)
        uiDlg.groupBox.setFocusPolicy(Qt.FocusPolicy.TabFocus)
        for i in range(len(buttons)-1):
            uiDlg.groupBox.setTabOrder(buttons[i], buttons[i+1])

    def tabWidgetChangedSlot(self, index):
        applyEnabled = False
        if index == 1:  # Custom Sequence
            self.validateCustomSequence()
            if self.customSequenceIsValid:
                applyEnabled = True
        else:  # Standard Sequence
            self.useCustomSequence = False
            if self.chosenStandardSequence != None:
                # Overwrite sequence in case custom has been applied
                activeButton = self.buttons[self.chosenStandardSequence]
                sequenceName = str(activeButton.text())
                self.validatedSequenceToApply = sequences.get(sequenceName, None)
                applyEnabled = True
        self.applyButton.setEnabled(applyEnabled)

    def standardSequenceChangedSlot(self, optionChosen):
        """
        Connected to signalMapper to receive a signal whenever user selects
        a different sequence in the standard tab.
        """
        sequenceName = str(self.buttons[optionChosen].text())
        self.validatedSequenceToApply = sequences.get(sequenceName, None)
        self.chosenStandardSequence = optionChosen
        self.applyButton.setEnabled(True)

    def validateCustomSequence(self):
        """
        Called when:
        1. User enters custom sequence (i.e. seqBox emits textChanged signal)
        2. tabWidgetChangedSlot sees the user has switched to custom tab.

        When the sequence is valid, make the applyButton active for clicking.
        Otherwise
        """
        userSequence = self.seqBox.toPlainText()
        if len(userSequence) == 0:
            self.customSequenceIsValid = False
            return  # tabWidgetChangedSlot will disable applyButton
        
        userMatch = dnapattern.match(userSequence)
        if not userMatch.hasMatch():  # no invalid characters
            self.useCustomSequence = True
            self.customSequenceIsValid = True
            self.applyButton.setEnabled(True)
        else:
            self.customSequenceIsValid = False
            self.applyButton.setEnabled(False)

    def applySequence(self, oligo):
        self.dialog.setFocus()
        if self.dialog.exec():  # apply the sequence if accept was clicked
            if self.useCustomSequence:
                self.validatedSequenceToApply = str(self.seqBox.toPlainText().upper())
            oligo.applySequence(self.validatedSequenceToApply)
            return oligo.length(), len(self.validatedSequenceToApply)
        return (None, None)



#!/usr/bin/env python
# encoding: utf-8

from operator import attrgetter
import cadnano2.util as util
from array import array

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QObject', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QUndoStack', 'QUndoCommand'])

class Decorator(object):
    """
    Decorators do not affect an applied sequence
    """
    def __init__(self, index):
        if self.__class__ == Decorator:
            e = "Decorator should be subclassed."
            raise NotImplementedError(e)
        self._index = idex
        self._dType = None
        self._privateSequence = None
    # end def

    def privateLength(self):
        """
        This is the length of a sequence that is immutable by the strand
        """
        return length(self._privateSequence)

    def decoratorType(self):
        return self._dtype
# end class


QObject = object

class Qt(object):
    pass

class pyqtSignal(object):
    def __init__(self, *args):
        """ We don't actually do anything with argtypes because
        the real Qt will perform checks in the Gui version of
        cadnano which should suffice. """
        self.signalEmitters = {}
        self.argtypes = args
    def __get__(self, emitter):
        try:
            return self.signalEmitters.get(emitter)
        except KeyError:
            self.signalEmitters[emitter] = pyqtBoundSignal()
            return emitter

class pyqtBoundSignal(object):
    def __init__(self):
        self.targets = {}
    def connect(self, target):
        self.targets.add(target)
    def disconnect(self, target):
        self.targets.remove(target)
    def emit(self, *args):
        for t in self.targets:
            t(*args)


#!/usr/bin/env python
# encoding: utf-8

import cadnano2.util as util
import copy
from .strand import Strand
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject'])
util.qtWrapImport('QtGui', globals(), ['QUndoCommand'])


class Oligo(QObject):
    """
    Oligo is a group of Strands that are connected via 5' and/or 3'
    connections. It corresponds to the physical DNA strand, and is thus
    used tracking and storing properties that are common to a single strand,
    such as its color.

    Commands that affect Strands (e.g. create, remove, merge, split) are also
    responsible for updating the affected Oligos.
    """
    def __init__(self, part, color=None):
        super(Oligo, self).__init__(part)
        self._part = part
        self._strand5p = None
        self._length = 0
        self._isLoop = False
        self._color = color if color else "#0066cc"
    # end def

    def __repr__(self):
        clsName = self.__class__.__name__
        olgId = str(id(self))[-4:]
        strandType = "Stap" if self.isStaple() else "Scaf"
        vhNum = self._strand5p.strandSet().virtualHelix().number()
        idx = self._strand5p.idx5Prime()
        return "<%s %s>(%s %d[%d])" % (clsName, olgId, strandType, vhNum, idx)

    def shallowCopy(self):
        olg = Oligo(self._part)
        olg._strand5p = self._strand5p
        olg._length = self._length
        olg._isLoop = self._isLoop
        olg._color = self._color
        return olg
    # end def

    def deepCopy(self, part):
        olg = Oligo(part)
        olg._strand5p = None
        olg._length = self._length
        olg._isLoop = self._isLoop
        olg._color = self._color
        return olg
    # end def

    ### SIGNALS ###
    oligoIdentityChangedSignal = pyqtSignal(QObject)  # new oligo
    oligoAppearanceChangedSignal = pyqtSignal(QObject)  # self
    oligoSequenceAddedSignal = pyqtSignal(QObject)  # self
    oligoSequenceClearedSignal = pyqtSignal(QObject)  # self

    ### SLOTS ###

    ### ACCESSORS ###
    def color(self):
        return self._color
    # end def

    def locString(self):
        vhNum = self._strand5p.strandSet().virtualHelix().number()
        idx = self._strand5p.idx5Prime()
        return "%d[%d]" % (vhNum, idx)
    # end def

    def part(self):
        return self._part
    # end def

    def strand5p(self):
        return self._strand5p
    # end def

    def setStrand5p(self, strand):
        self._strand5p = strand
    # end def

    def undoStack(self):
        return self._part.undoStack()
    # end def

    ### PUBLIC METHODS FOR QUERYING THE MODEL ###
    def isLoop(self):
        return self._isLoop

    def isStaple(self):
        return self._strand5p.isStaple()

    def length(self):
        return self._length
    # end def

    def sequence(self):
        temp = self.strand5p()
        if not temp:
            return None
        if temp.sequence():
            return ''.join([Strand.sequence(strand) \
                        for strand in self.strand5p().generator3pStrand()])
        else:
            return None
    # end def

    def sequenceExport(self):
        vhNum5p = self.strand5p().virtualHelix().number()
        idx5p = self.strand5p().idx5Prime()
        seq = ''
        if self.isLoop():
            # print "A loop exists"
            raise Exception
        for strand in self.strand5p().generator3pStrand():
            seq = seq + Strand.sequence(strand, forExport=True)
            if strand.connection3p() == None:  # last strand in the oligo
                vhNum3p = strand.virtualHelix().number()
                idx3p = strand.idx3Prime()
        output = "%d[%d],%d[%d],%s,%s,%s\n" % \
                (vhNum5p, idx5p, vhNum3p, idx3p, seq, len(seq), self._color)
        return output
    # end def

    def shouldHighlight(self):
        if not self._strand5p:
            return False
        if self._strand5p.isScaffold():
            return False
        if self.length() < 18:
            return True
        if self.length() > 50:
            return True
        return False
    # end def

    ### PUBLIC METHODS FOR EDITING THE MODEL ###
    def remove(self, useUndoStack=True):
        c = Oligo.RemoveOligoCommand(self)
        util.execCommandList(self, [c], desc="Color Oligo", useUndoStack=useUndoStack)
    # end def

    def applyColor(self, color, useUndoStack=True):
        if color == self._color:
            return  # oligo already has color
        c = Oligo.ApplyColorCommand(self, color)
        util.execCommandList(self, [c], desc="Color Oligo", useUndoStack=useUndoStack)
    # end def

    def applySequence(self, sequence, useUndoStack=True):
        c = Oligo.ApplySequenceCommand(self, sequence)
        util.execCommandList(self, [c], desc="Apply Sequence", useUndoStack=useUndoStack)
    # end def

    def applySequenceCMD(self, sequence, useUndoStack=True):
        return Oligo.ApplySequenceCommand(self, sequence)
    # end def

    def setLoop(self, bool):
        self._isLoop = bool

    ### PUBLIC SUPPORT METHODS ###
    def addToPart(self, part):
        self._part = part
        self.setParent(part)
        part.addOligo(self)
    # end def

    def destroy(self):
        # QObject also emits a destroyed() Signal
        self.setParent(None)
        self.deleteLater()
    # end def

    def decrementLength(self, delta):
        self.setLength(self._length-delta)
    # end def

    def incrementLength(self, delta):
        self.setLength(self._length+delta)
    # end def

    def refreshLength(self):
        temp = self.strand5p()
        if not temp:
            return
        length = 0
        for strand in temp.generator3pStrand():
            length += strand.totalLength()
        self.setLength(length)
    # end def

    def removeFromPart(self):
        """
        This method merely disconnects the object from the model.
        It still lives on in the undoStack until clobbered

        Note: don't set self._part = None because we need to continue passing
        the same reference around.
        """
        self._part.removeOligo(self)
        self.setParent(None)
    # end def

    def setColor(self, color):
        self._color = color
    # end def

    def setLength(self, length):
        before = self.shouldHighlight()
        self._length = length
        if before != self.shouldHighlight():
            self.oligoSequenceClearedSignal.emit(self)
            self.oligoAppearanceChangedSignal.emit(self)
    # end def

    def strandMergeUpdate(self, oldStrandLow, oldStrandHigh, newStrand):
        """
        This method sets the isLoop status of the oligo and the oligo's
        5' strand.
        """
        # check loop status
        if oldStrandLow.oligo() == oldStrandHigh.oligo():
            self._isLoop = True
            self._strand5p = newStrand
            return 
            # leave the _strand5p as is?
        # end if

        # Now get correct 5p end to oligo
        if oldStrandLow.isDrawn5to3():
            if oldStrandLow.connection5p() != None:
                self._strand5p = oldStrandLow.oligo()._strand5p
            else:
                self._strand5p = newStrand
        else:
            if oldStrandHigh.connection5p() != None:
                self._strand5p = oldStrandHigh.oligo()._strand5p
            else:
                self._strand5p = newStrand
        # end if
    # end def

    def strandResized(self, delta):
        """
        Called by a strand after resize. Delta is used to update the length,
        which may case an appearance change.
        """
        pass
    # end def

    def strandSplitUpdate(self, newStrand5p, newStrand3p, oligo3p, oldMergedStrand):
        """
        If the oligo is a loop, splitting the strand does nothing. If the
        oligo isn't a loop, a new oligo must be created and assigned to the
        newStrand and everything connected to it downstream.
        """
        # if you split it can't be a loop
        self._isLoop = False
        if oldMergedStrand.oligo().isLoop():
            self._strand5p = newStrand3p
            return
        else:
            if oldMergedStrand.connection5p() == None:
                self._strand5p = newStrand5p
            else:
                self._strand5p = oldMergedStrand.oligo()._strand5p
            oligo3p._strand5p = newStrand3p
        # end else
    # end def

    ### PRIVATE SUPPORT METHODS ###

    ### COMMANDS ###
    class ApplyColorCommand(QUndoCommand):
        def __init__(self, oligo, color):
            super(Oligo.ApplyColorCommand, self).__init__()
            self._oligo = oligo
            self._newColor = color
            self._oldColor = oligo.color()
        # end def

        def redo(self):
            olg = self._oligo
            olg.setColor(self._newColor)
            olg.oligoAppearanceChangedSignal.emit(olg)
        # end def

        def undo(self):
            olg = self._oligo
            olg.setColor(self._oldColor)
            olg.oligoAppearanceChangedSignal.emit(olg)
        # end def
    # end class

    class ApplySequenceCommand(QUndoCommand):
        def __init__(self, oligo, sequence):
            super(Oligo.ApplySequenceCommand, self).__init__()
            self._oligo = oligo
            self._newSequence = sequence
            self._oldSequence = oligo.sequence()
            self._strandType = oligo._strand5p.strandSet().strandType()
        # end def

        def redo(self):
            olg = self._oligo
            nS = ''.join(self._newSequence) if self._newSequence else None
            nS_original = self._newSequence
            oligoList = [olg]
            for strand in olg.strand5p().generator3pStrand():
                usedSeq, nS = strand.setSequence(nS)
                # get the compliment ahead of time
                usedSeq = util.comp(usedSeq) if usedSeq else None
                compSS = strand.strandSet().complementStrandSet()
                for compStrand in compSS._findOverlappingRanges(strand):
                    subUsedSeq = compStrand.setComplementSequence(usedSeq, strand)
                    oligoList.append(compStrand.oligo())
                # end for
                # as long as the new Applied Sequence is not None
                if nS == None and nS_original:
                    break
            # end for
            for oligo in oligoList:
                oligo.oligoSequenceAddedSignal.emit(oligo)
        # end def

        def undo(self):
            olg = self._oligo
            oS = ''.join(self._oldSequence) if self._oldSequence else None

            oligoList = [olg]

            for strand in olg.strand5p().generator3pStrand():
                usedSeq, oS = strand.setSequence(oS)

                # get the compliment ahead of time
                usedSeq = util.comp(usedSeq) if usedSeq else None
                compSS = strand.strandSet().complementStrandSet()
                for compStrand in compSS._findOverlappingRanges(strand):
                    subUsedSeq = compStrand.setComplementSequence(usedSeq, strand)
                    oligoList.append(compStrand.oligo())
                # end for
            # for

            for oligo in oligoList:
                oligo.oligoSequenceAddedSignal.emit(oligo)
        # end def
    # end class
    class ApplyColorCommand(QUndoCommand):
        def __init__(self, oligo, color):
            super(Oligo.ApplyColorCommand, self).__init__()
            self._oligo = oligo
            self._newColor = color
            self._oldColor = oligo.color()
        # end def

        def redo(self):
            olg = self._oligo
            olg.setColor(self._newColor)
            olg.oligoAppearanceChangedSignal.emit(olg)
        # end def

        def undo(self):
            olg = self._oligo
            olg.setColor(self._oldColor)
            olg.oligoAppearanceChangedSignal.emit(olg)
        # end def
    # end class

    class RemoveOligoCommand(QUndoCommand):
        def __init__(self,oligo):
            super(Oligo.RemoveOligoCommand, self).__init__()
            self._oligo = oligo
            self._part = oligo.part()
            self._strandIdxList = []
            self._strand3p = None
        # end def

        def redo(self):
            sIList = self._strandIdxList
            o = self._oligo
            s5p = o.strand5p()
            part = self._part

            for strand in s5p.generator3pStrand():
                strandSet = strand.strandSet()
                strandSet._doc.removeStrandFromSelection(strand)
                isInSet, overlap, sSetIdx = strandSet._findIndexOfRangeFor(strand)
                sIList.append(sSetIdx)
                strandSet._strandList.pop(sSetIdx)
                # Emit a signal to notify on completion
                strand.strandRemovedSignal.emit(strand)
                # for updating the Slice View displayed helices
                strandSet.part().partStrandChangedSignal.emit(strandSet.part(), strandSet.virtualHelix())
            # end def
            # set the 3p strand for the undo
            self._strand3p = strand

            # remove Oligo from part but don't set parent to None?
            # o.removeFromPart()
            part.removeOligo(o)
        # end def

        def undo(self):
            sIList = self._strandIdxList
            o = self._oligo
            s3p = self._strand3p
            part = self._part

            for strand in s3p.generator5pStrand():
                strandSet = strand.strandSet()
                sSetIdx = sIList.pop(-1)
                strandSet._strandList.insert(sSetIdx, strand)
                # Emit a signal to notify on completion
                strandSet.strandsetStrandAddedSignal.emit(strandSet, strand)
                # for updating the Slice View displayed helices
                part.partStrandChangedSignal.emit(strandSet.part(), strandSet.virtualHelix())
            # end def

            # add Oligo to part but don't set parent to None?
            # o.addToPart(part)
            part.addOligo(o)
        # end def
    # end class
# end class
    


"""
modeltests.py

Created by Shawn Douglas on 2011-06-28.

Instructions: copypaste the following script into TextMate's
Bundles > Bundle Editor > Show Bundle Editor > Python > Run Project Unit Tests

cd "$TM_PROJECT_DIRECTORY"
find . -name "*tests.py" -exec "${TM_PYTHON:-python}" '{}' \;|pre

"""

import sys
sys.path.insert(0, '.')

import tests.cadnanoguitestcase
from tests.cadnanoguitestcase import CadnanoGuiTestCase
import time
from cadnano2.model.virtualhelix import VirtualHelix
from cadnano2.model.enum import StrandType


class ModelTests(CadnanoGuiTestCase):
    """
    Create new tests by adding methods to this class that begin with "test".
    See for more detail: http://docs.python.org/library/unittest.html

    Run model tests by calling "python -m tests.modeltests" from cadnano2 root
    directory.
    """
    def setUp(self):
        """
        The setUp method is called before running any test. It is used
        to set the general conditions for the tests to run correctly.
        """
        CadnanoGuiTestCase.setUp(self)
        # Add extra model-test-specific initialization here

    def tearDown(self):
        """
        The tearDown method is called at the end of running each test,
        generally used to clean up any objects created in setUp
        """
        CadnanoGuiTestCase.tearDown(self)
        # Add model-test-specific cleanup here

    def testModel1(self):
        """docstring for testModel1"""
        pass


if __name__ == '__main__':
    print("Running Model Tests")
    tests.cadnanoguitestcase.main()



from cadnano2.controllers.itemcontrollers.partitemcontroller import PartItemController
from .emptyhelixitem import EmptyHelixItem
from .virtualhelixitem import VirtualHelixItem
from .activesliceitem import ActiveSliceItem
from cadnano2.views import styles
import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QRectF', 'QPointF', 'QEvent', 'Qt',
                                        'pyqtSignal', 'pyqtSlot', 'QObject'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QPainterPath',
                                       'QPen'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem',
                                           'QGraphicsEllipseItem'])

_radius = styles.SLICE_HELIX_RADIUS
_defaultRect = QRectF(0, 0, 2 * _radius, 2 * _radius)
highlightWidth = styles.SLICE_HELIX_MOD_HILIGHT_WIDTH
delta = (highlightWidth - styles.SLICE_HELIX_STROKE_WIDTH)/2.
_hoverRect = _defaultRect.adjusted(-delta, -delta, delta, delta)
_modPen = QPen(styles.bluestroke, highlightWidth)


class PartItem(QGraphicsItem):
    _radius = styles.SLICE_HELIX_RADIUS

    def __init__(self, modelPart, parent=None):
        """
        Parent should be either a SliceRootItem, or an AssemblyItem.

        Invariant: keys in _emptyhelixhash = range(_nrows) x range(_ncols)
        where x is the cartesian product.

        Order matters for deselector, probe, and setlattice
        """
        super(PartItem, self).__init__(parent)
        self._part = modelPart
        self._controller = PartItemController(self, modelPart)
        self._activeSliceItem = ActiveSliceItem(self, modelPart.activeBaseIndex())
        self._scaleFactor = self._radius/modelPart.radius()
        self._emptyhelixhash = {}
        self._virtualHelixHash = {}
        self._nrows, self._ncols = 0, 0
        self._rect = QRectF(0, 0, 0, 0)
        self._initDeselector()
        # Cache of VHs that were active as of last call to activeSliceChanged
        # If None, all slices will be redrawn and the cache will be filled.
        # Connect destructor. This is for removing a part from scenes.
        self.probe = self.IntersectionProbe(self)
        # initialize the PartItem with an empty set of old coords
        self._setLattice([], modelPart.generatorFullLattice())
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemHasNoContents)  # never call paint
        self.setZValue(styles.ZPARTITEM)
        self._initModifierCircle()
    # end def

    def _initDeselector(self):
        """
        The deselector grabs mouse events that missed a slice and clears the
        selection when it gets one.
        """
        self.deselector = ds = PartItem.Deselector(self)
        ds.setParentItem(self)
        ds.setFlag(QGraphicsItem.GraphicsItemFlag.ItemStacksBehindParent)
        ds.setZValue(styles.ZDESELECTOR)

    def _initModifierCircle(self):
        self._canShowModCirc = False
        self._modCirc = mC = QGraphicsEllipseItem(_hoverRect, self)
        mC.setPen(_modPen)
        mC.hide()
    # end def

    ### SIGNALS ###

    ### SLOTS ###
    def partActiveVirtualHelixChangedSlot(self, part, virtualHelix):
        pass

    def partDimensionsChangedSlot(self, sender):
        pass
    # end def

    def partHideSlot(self, sender):
        self.hide()
    # end def

    def partParentChangedSlot(self, sender):
        """docstring for partParentChangedSlot"""
        # print "PartItem.partParentChangedSlot"
        pass

    def partRemovedSlot(self, sender):
        """docstring for partRemovedSlot"""
        self._activeSliceItem.removed()
        self.parentItem().removePartItem(self)

        scene = self.scene()

        self._virtualHelixHash = None

        for item in list(self._emptyhelixhash.items()):
            key, val = item
            scene.removeItem(val)
            del self._emptyhelixhash[key]
        self._emptyhelixhash = None

        scene.removeItem(self)

        self._part = None
        self.probe = None
        self._modCirc = None

        self.deselector = None
        self._controller.disconnectSignals()
        self._controller = None
    # end def

    def partVirtualHelicesReorderedSlot(self, sender, orderedCoordList):
        pass
    # end def

    def partPreDecoratorSelectedSlot(self, sender, row, col, baseIdx):
        """docstring for partPreDecoratorSelectedSlot"""
        vhi = self.getVirtualHelixItemByCoord(row, col)
        view = self.window().sliceGraphicsView
        view.sceneRootItem.resetTransform()
        view.centerOn(vhi)
        view.zoomIn()
        mC = self._modCirc
        x,y = self._part.latticeCoordToPositionXY(row, col, self.scaleFactor())
        mC.setPos(x,y)
        if self._canShowModCirc:
            mC.show()
    # end def

    def partVirtualHelixAddedSlot(self, sender, virtualHelix):
        vh = virtualHelix
        coords = vh.coord()

        emptyHelixItem = self._emptyhelixhash[coords]
        # TODO test to see if self._virtualHelixHash is necessary
        vhi = VirtualHelixItem(vh, emptyHelixItem)
        self._virtualHelixHash[coords] = vhi
    # end def

    def partVirtualHelixRenumberedSlot(self, sender, coord):
        pass
    # end def

    def partVirtualHelixResizedSlot(self, sender, coord):
        pass
    # end def

    def updatePreXoverItemsSlot(self, sender, virtualHelix):
        pass
    # end def

    ### ACCESSORS ###
    def boundingRect(self):
        return self._rect
    # end def

    def part(self):
        return self._part
    # end def

    def scaleFactor(self):
        return self._scaleFactor
    # end def

    def setPart(self, newPart):
        self._part = newPart
    # end def

    def window(self):
        return self.parentItem().window()
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _upperLeftCornerForCoords(self, row, col):
        pass  # subclass
    # end def

    def _updateGeometry(self):
        self._rect = QRectF(0, 0, *self.part().dimensions())
    # end def

    def _spawnEmptyHelixItemAt(self, row, column):
        helix = EmptyHelixItem(row, column, self)
        # helix.setFlag(QGraphicsItem.GraphicsItemFlag.ItemStacksBehindParent, True)
        self._emptyhelixhash[(row, column)] = helix
    # end def

    def _killHelixItemAt(row, column):
        s = self._emptyhelixhash[(row, column)]
        s.scene().removeItem(s)
        del self._emptyhelixhash[(row, column)]
    # end def

    def _setLattice(self, oldCoords, newCoords):
        """A private method used to change the number of rows,
        cols in response to a change in the dimensions of the
        part represented by the receiver"""
        oldSet = set(oldCoords)
        oldList = list(oldSet)
        newSet = set(newCoords)
        newList = list(newSet)
        for coord in oldList:
            if coord not in newSet:
                self._killHelixItemAt(*coord)
        # end for
        for coord in newList:
            if coord not in oldSet:
                self._spawnEmptyHelixItemAt(*coord)
        # end for
        # self._updateGeometry(newCols, newRows)
        # self.prepareGeometryChange()
        # the Deselector copies our rect so it changes too
        self.deselector.prepareGeometryChange()
        self.zoomToFit()
    # end def

    ### PUBLIC SUPPORT METHODS ###
    def getVirtualHelixItemByCoord(self, row, column):
        if (row, column) in self._emptyhelixhash:
            return self._virtualHelixHash[(row, column)]
        else:
            return None
    # end def

    def paint(self, painter, option, widget=None):
        pass
    # end def

    def selectionWillChange(self, newSel):
        if self.part() == None:
            return
        if self.part().selectAllBehavior():
            return
        for sh in self._emptyhelixhash.values():
            sh.setSelected(sh.virtualHelix() in newSel)
    # end def

    def setModifyState(self, bool):
        """Hides the modRect when modify state disabled."""
        self._canShowModCirc = bool
        if bool == False:
            self._modCirc.hide()

    def updateStatusBar(self, statusString):
        """Shows statusString in the MainWindow's status bar."""
        pass  # disabled for now.
        # self.window().statusBar().showMessage(statusString, timeout)

    def vhAtCoordsChanged(self, row, col):
        self._emptyhelixhash[(row, col)].update()
    # end def

    def zoomToFit(self):
        thescene = self.scene()
        theview = thescene.views()[0]
        theview.zoomToFit()
    # end def

    ### EVENT HANDLERS ###
    def mousePressEvent(self, event):
        # self.createOrAddBasesToVirtualHelix()
        QGraphicsItem.mousePressEvent(self, event)
    # end def

    class Deselector(QGraphicsItem):
        """The deselector lives behind all the slices and observes mouse press
        events that miss slices, emptying the selection when they do"""
        def __init__(self, parentHGI):
            super(PartItem.Deselector, self).__init__()
            self.parentHGI = parentHGI

        def mousePressEvent(self, event):
            self.parentHGI.part().setSelection(())
            super(PartItem.Deselector, self).mousePressEvent(event)

        def boundingRect(self):
            return self.parentHGI.boundingRect()

        def paint(self, painter, option, widget=None):
            pass

    class IntersectionProbe(QGraphicsItem):
        def boundingRect(self):
            return QRectF(0, 0, .1, .1)
        def paint(self, painter, option, widget=None):
            pass
